import pytest
from unittest import mock

from pymatrix import pymatrix


def test_init():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        with mock.patch.object(pymatrix.random, "choice", return_value="X"):
            test_line = pymatrix.OldScrollingLine(5, 10, 10)
            assert test_line.height == 8
            assert test_line.width == 9
            assert test_line.y == -1
            assert test_line.x == 5
            assert test_line.length == 3
            assert test_line.lead_y == 0
            assert test_line.lead_char == "X"
            assert test_line.location_list == []
            assert test_line.line_color_number == 3


@pytest.mark.parametrize("value, expected", [
    (1, True), (2, False), (3, False),
])
def test_init_bold(value, expected):
    with mock.patch.object(pymatrix.random,"choice", return_value=["X"]):
        with mock.patch.object(pymatrix.random, "randint", return_value=value):
            test_line = pymatrix.OldScrollingLine(5, 10, 10)
            assert test_line.bold == expected


def test_old_school_char_list_start():
    assert pymatrix.OldScrollingLine.old_scroll_chr_list == []


def test_update_char_list():
    pymatrix.OldScrollingLine.update_char_list(["t", "1"])
    assert pymatrix.OldScrollingLine.old_scroll_chr_list == ["t", "1"]


def test_get_lead_first():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            lead = test_line.get_lead()
            assert lead == (0, 5, "X")


def test_get_lead_second():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            _ = test_line.get_lead()
            lead = test_line.get_lead()
            assert lead == (1, 5, "X")


def test_get_lead_last():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            test_line.lead_y = 3
            lead = test_line.get_lead()
            assert lead == (3, 5, "X")


def test_get_lead_off_screen():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            test_line.lead_y = 4
            lead = test_line.get_lead()
            assert lead is None


def test_get_next_first():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            line_list = test_line.get_next()
            assert line_list == []


def test_get_next_second():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            _ = test_line.get_next()
            line_list = test_line.get_next()
            assert line_list == [[0, 5, "X"]]


def test_get_next_third():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            for x in range(2):
                _ = test_line.get_next()
            line_list = test_line.get_next()
            assert line_list == [[1, 5, "X"], [0, 5, "X"]]


def test_get_next_line_length():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            for x in range(4):
                _ = test_line.get_next()
            line_list = test_line.get_next()
            assert line_list == [[3, 5, "X"], [2, 5, "X"], [1, 5, "X"]]


def test_get_next_off_screen():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            for x in range(5):
                _ = test_line.get_next()
            line_list = test_line.get_next()
            assert line_list == [[3, 5, "X"], [2, 5, "X"]]


def test_delete_last_first():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            remove = test_line.delete_last()
            assert remove is None


def test_delete_last_line_length():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            for x in range(4):
                _ = test_line.get_next()
            remove = test_line.delete_last()
            assert remove == [0, 5]


def test_delete_last_line_length_next():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            for x in range(5):
                _ = test_line.get_next()
            remove = test_line.delete_last()
            assert remove == [1, 5]


def test_okay_to_delete_first():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is False


def test_okay_to_delete_done():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 5)
            test_line.location_list = []
            test_line.y = 4
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is True


def test_full_line_run():
    with mock.patch.object(pymatrix.random, "choice", return_value="X"):
        with mock.patch.object(pymatrix.random, "randint", return_value=3):
            test_line = pymatrix.OldScrollingLine(5, 10, 6)
            # round 1
            remove = test_line.delete_last()
            assert remove is None
            lead = test_line.get_lead()
            assert lead == (0, 5, "X")
            loc_list = test_line.get_next()
            assert loc_list == []
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is False
            # round 2
            remove = test_line.delete_last()
            assert remove is None
            lead = test_line.get_lead()
            assert lead == (1, 5, "X")
            loc_list = test_line.get_next()
            assert loc_list == [[0, 5, "X"]]
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is False
            # round 3
            remove = test_line.delete_last()
            assert remove is None
            lead = test_line.get_lead()
            assert lead == (2, 5, "X")
            loc_list = test_line.get_next()
            assert loc_list == [[1, 5, "X"], [0, 5, "X"]]
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is False
            # round 4
            remove = test_line.delete_last()
            assert remove is None
            lead = test_line.get_lead()
            assert lead == (3, 5, "X")
            loc_list = test_line.get_next()
            assert loc_list == [[2, 5, "X"], [1, 5, "X"], [0, 5, "X"]]
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is False
            # round 5
            remove = test_line.delete_last()
            assert remove == [0, 5]
            lead = test_line.get_lead()
            assert lead == (4, 5, "X")
            loc_list = test_line.get_next()
            assert loc_list == [[3, 5, "X"], [2, 5, "X"], [1, 5, "X"]]
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is False
            # round 6
            remove = test_line.delete_last()
            assert remove == [1, 5]
            lead = test_line.get_lead()
            assert lead is None
            loc_list = test_line.get_next()
            assert loc_list == [[4, 5, "X"], [3, 5, "X"], [2, 5, "X"]]
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is False
            # round 7
            remove = test_line.delete_last()
            assert remove == [2, 5]
            lead = test_line.get_lead()
            assert lead is None
            loc_list = test_line.get_next()
            assert loc_list == [[4, 5, "X"], [3, 5, "X"]]
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is False
            # round 8
            remove = test_line.delete_last()
            assert remove == [3, 5]
            lead = test_line.get_lead()
            assert lead is None
            loc_list = test_line.get_next()
            assert loc_list == [[4, 5, "X"]]
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is False
            # round 9
            remove = test_line.delete_last()
            assert remove == [4, 5]
            lead = test_line.get_lead()
            assert lead is None
            loc_list = test_line.get_next()
            assert loc_list == []
            okay_to_delete = test_line.okay_to_delete()
            assert okay_to_delete is True



#! /usr/bin/python3
""" Matrix style rain using Python 3 and curses. """
import argparse
import curses
import datetime
import importlib.metadata
import itertools
import os
import random
import time

from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union


version = importlib.metadata.version("pymatrix-rain")

CHAR_LIST = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m",
             "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z",
             "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M",
             "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z",
             "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "!", "#", "$",
             "%", "^", "&", "(", ")", "-", "+", "=", "[", "]", "{", "}", "|",
             ";", ":", "<", ">", ",", ".", "?", "~", "`", "@", "*", "_", "'",
             "\\", "/", '"']

EXT_CHAR_LIST = ["Ç", "È", "Ì", "Í", "Ð", "Ñ", "Ò", "×", "Ø", "Ù", "Ú", "Ý",
                 "Þ", "ß", "à", "£", "¤", "¥", "§", "ª", "¶", "º", "»", "¿",
                 "Ä", "Å", "é", "ê", "í", "ï", "å", "æ", "ç", "è", "ð", "ñ",
                 "ò", "ö", "ø", "ù", "ý", "þ", "ā", "ć", "ĉ", "ė", "ě", "ĝ",
                 "ģ", "ħ", "ī", "ı", "ķ", "Ľ", "Ł", "ł", "ń", "ň", "ō", "Œ",
                 "œ", "ŕ", "ŗ", "ś", "ŝ", "š", "ť", "ū", "ų", "Ÿ", "ź", "ż",
                 "Ž", "ž", "ș", "ț", "ë", "Ĉ", "Ď", "ď", "Ġ", "Ř", "°", "«",
                 "±", "Δ", "Ξ", "Λ", ]

KATAKANA_CHAR_LIST = ["ｦ", "ｱ", "ｳ", "ｴ", "ｵ", "ｶ", "ｷ", "ｹ", "ｺ", "ｻ", "ｼ",
                      "ｽ", "ｾ", "ｿ", "ﾀ", "ﾂ", "ﾃ", "ﾅ", "ﾆ", "ﾇ", "ﾈ", "ﾊ",
                      "ﾋ", "ﾎ", "ﾏ", "ﾐ", "ﾑ", "ﾒ", "ﾓ", "ﾔ", "ﾕ", "ﾗ", "ﾘ",
                      "ﾜ", "ﾍ", "ｲ", "ｸ", "ﾁ", "ﾄ", "ﾉ", "ﾌ", "ﾖ", "ﾙ", "ﾚ",
                      "ﾛ", "ﾝ"]

KATAKANA_CHAR_LIST_ADDON = ["0", "1", "2", "3", "4", "5", "7", "8", "9", "Z",
                            ":", ".", "=", "*", "+", "-", "<", ">"]


DELAY_SPEED = {0: 0.005, 1: 0.01, 2: 0.025, 3: 0.04, 4: 0.055, 5: 0.07,
               6: 0.085, 7: 0.1, 8: 0.115, 9: 0.13}

CURSES_CH_CODES_CYCLE_DELAY = {41: 1, 33: 2, 64: 3, 35: 4, 36: 5, 37: 6,
                               94: 7, 38: 8, 42: 9, 40: 10}

CURSES_COLOR = {"red": curses.COLOR_RED, "green": curses.COLOR_GREEN,
                "blue": curses.COLOR_BLUE, "yellow": curses.COLOR_YELLOW,
                "magenta": curses.COLOR_MAGENTA, "cyan": curses.COLOR_CYAN,
                "white": curses.COLOR_WHITE, "black": curses.COLOR_BLACK}

CURSES_OVER_RIDE_COLORS = {"red": 160, "green": 40, "blue": 21,
                           "yellow": 184, "magenta": 164, "cyan": 44,
                           "white": 255, "black": 16}

CURSES_CH_CODES_COLOR = {114: "red", 82: "red", 116: "green", 84: "green",
                         121: "blue", 89: "blue", 117: "yellow", 85: "yellow",
                         105: "magenta", 73: "magenta", 111: "cyan",
                         79: "cyan", 112: "white", 80: "white", 18: "red",
                         20: "green", 25: "blue", 21: "yellow", 9: "magenta",
                         15: "cyan", 16: "white", 27: "black", 91: "black",
                         123: "black"}
DEFAULT_BG_CHAR = " "
DEFAULT_CYCLE_COLOR_DELAY = 500
WAKE_UP_PAIR = 21
WAKE_UP_KEYS = [119, 65, 107, 101]
MIN_SCREEN_SIZE_Y = 10
MIN_SCREEN_SIZE_X = 10


class PyMatrixError(Exception):
    pass


class SingleLine:
    def __init__(self, y: int, x: int, width: int, height: int, direction: str):
        self.direction = direction
        self.height = height - 2
        self.width = width - 1
        self.async_scroll_count = 0
        self.async_scroll_rate = random.randint(0, 4)
        self.line_color_number = random.randint(1, 7)  # keep for now
        if direction == "down":
            self.lead_y = 0
            self.y = -1
            self.x = x
            length = random.randint(3, height - 3)
            self.last_y = -length  # track when to start removing characters
        elif direction == "up":
            self.lead_y = height - 2
            self.y = height - 1
            self.x = x
            length = random.randint(3, height - 3)
            self.last_y = height - 3 + length
        elif direction == "right":
            self.lead_x = 0
            self.x = -1
            length = random.randint(3, width - 3)
            self.last_x = -length
            self.y = y
            self.lead_y = 0
            self.last_y = 0
        elif direction == "left":
            self.lead_x = width - 2
            self.x = width - 1
            length = random.randint(3, width - 3)
            self.last_x = width - 2 + length
            self.y = y
            self.lead_y = 0
            self.last_y = 0

    def get_lead(self) -> Union[Tuple[int, int], None]:
        if self.direction == "down":
            if self.lead_y > self.height:
                return None
            else:
                lead_y = self.lead_y
                self.lead_y += 1
                return lead_y, self.x
        elif self.direction == "up":
            if self.lead_y < 0:
                return None
            else:
                lead_y = self.lead_y
                self.lead_y -= 1
                return lead_y, self.x
        elif self.direction == "right":
            if self.lead_x >= self.width:
                return None
            else:
                lead_x = self.lead_x
                self.lead_x += 1
                return self.y, lead_x
        elif self.direction == "left":
            if self.lead_x < 0:
                return None
            else:
                lead_x = self.lead_x
                self.lead_x -= 1
                return self.y, lead_x

    def get_next(self) -> Union[Tuple[int, int], None]:
        if self.direction == "down":
            if self.y < 0 or self.y > self.height:
                self.y += 1
                return None
            else:
                y = self.y
                self.y += 1
                return y, self.x
        elif self.direction == "up":
            if self.y > self.height or self.y < 0:
                self.y -= 1
                return None
            else:
                y = self.y
                self.y -= 1
                return y, self.x
        elif self.direction == "right":
            if self.x < 0 or self.x >= self.width:
                self.x += 1
                return None
            else:
                x = self.x
                self.x += 1
                return self.y, x
        elif self.direction == "left":
            if self.x >= self.width or self.x < 0:
                self.x -= 1
                return None
            else:
                x = self.x
                self.x -= 1
                return self.y, x

    def delete_last(self) -> Union[Tuple[int, int], None]:
        if self.direction == "down":
            if self.last_y < 0 or self.last_y > self.height:
                self.last_y += 1
                return None
            else:
                last_y = self.last_y
                self.last_y += 1
                return last_y, self.x
        elif self.direction == "up":
            if self.last_y > self.height or self.last_y < 0:
                self.last_y -= 1
                return None
            else:
                last_y = self.last_y
                self.last_y -= 1
                return last_y, self.x
        elif self.direction == "right":
            if self.last_x < 0 or self.last_x >= self.width:
                self.last_x += 1
                return None
            else:
                last_x = self.last_x
                self.last_x += 1
                return self.y, last_x
        elif self.direction == "left":
            if self.last_x >= self.width:
                self.last_x -= 1
                return None
            else:
                last_x = self.last_x
                self.last_x -= 1
                return self.y, last_x

    def okay_to_delete(self) -> bool:
        if self.direction == "down":
            return True if self.last_y > self.height else False
        elif self.direction == "up":
            return True if self.last_y < 0 else False
        elif self.direction == "right":
            return True if self.last_x >= self.width else False
        elif self.direction == "left":
            return True if self.last_x < 0 else False

    def async_scroll_turn(self) -> bool:
        if self.async_scroll_count == self.async_scroll_rate:
            self.async_scroll_count = 0
            return True
        else:
            self.async_scroll_count += 1
            return False


class OldScrollingLine:
    old_scroll_chr_list = []

    def __init__(self, x: int, width: int, height: int):
        self.height = height - 2
        self.width = width - 1
        self.y = -1
        self.x = x
        self.length = random.randint(3, height - 3)
        self.lead_y = 0
        self.lead_char = random.choice(OldScrollingLine.old_scroll_chr_list)
        self.location_list = []
        self.line_color_number = random.randint(1, 7)
        self.bold = True if random.randint(1, 3) <= 1 else False

    @classmethod
    def update_char_list(cls, updated_char_list: List[str]) -> None:
        OldScrollingLine.old_scroll_chr_list = updated_char_list

    def delete_last(self) -> Union[None, List[int]]:
        if len(self.location_list) == self.length or self.y >= self.length:
            return self.location_list[-1][0:2]
        else:
            return None

    def get_lead(self) -> Union[Tuple[int, int, str], None]:
        if self.lead_y <= self.height:
            lead_y = self.lead_y
            self.lead_y += 1
            return lead_y, self.x, self.lead_char
        else:
            return None

    def get_next(self) -> List[List]:
        if len(self.location_list) != 0 and self.y >= 0:
            for cell in self.location_list:
                cell[0] += 1
        if len(self.location_list) < self.length and 0 <= self.y < self.height:
            self.location_list.append(
                [0, self.x, random.choice(OldScrollingLine.old_scroll_chr_list)]
            )
        if self.y > self.height:
            self.location_list.pop(0)
        self.y += 1
        return self.location_list

    def okay_to_delete(self) -> bool:
        return len(self.location_list) == 0 and self.y > self.height


class Matrix:
    def __init__(self, screen, args: argparse.Namespace):
        self.screen = screen
        self.args = args
        curses.curs_set(0)  # Set the cursor to off.
        self.screen.timeout(0)  # Turn blocking off for screen.getch().
        self.setup_colors()
        self.screen.bkgd(self.args.bg_char, curses.color_pair(1))
        self.color_cycle = itertools.cycle([1, 2, 3, 4, 5, 6])
        self.color_cycle_count = itertools.count(start=0, step=1)
        self.color_cycle_delay = DEFAULT_CYCLE_COLOR_DELAY
        self.wake_up_time = 20 if self.args.test_mode \
            else random.randint(2000, 3000)
        self.char_set = build_character_set2(args)
        if args.reverse:
            self.dir = "up"
        elif args.scroll_right:
            self.dir = "right"
        elif args.scroll_left:
            self.dir = "left"
        elif args.old_school_scrolling:
            self.dir = "old scrolling"
        else:
            self.dir = "down"
        if self.args.multiple_mode:
            self.color_mode = "multiple"
            setup_curses_colors("random",
                                self.args.background, self.args.over_ride)
        elif self.args.random_mode:
            self.color_mode = "random"
            setup_curses_colors("random",
                                self.args.background, self.args.over_ride)
        elif self.args.cycle:
            self.color_mode = "cycle"
        else:
            self.color_mode = "normal"
        self.spacer = 2 if self.args.double_space else 1
        self.line_list = []
        self.x_list = []
        self.y_list = []
        self.keys_pressed = []
        self.main_loop()

    def main_loop(self) -> None:
        size_y, size_x = self.screen.getmaxyx()
        self.check_screen_size(size_y, size_x)
        self.x_list = [x for x in range(0, size_x, self.spacer)]
        self.y_list = [y for y in range(1, size_y)]

        time_delta = datetime.timedelta(seconds=self.args.run_timer)
        end_time = datetime.datetime.now() + time_delta
        while True:
            if curses.is_term_resized(size_y, size_x):
                size_y, size_x = self.screen.getmaxyx()
                self.check_screen_size(size_y, size_x)
                self.x_list = [x for x in range(0, size_x, self.spacer)]
                self.y_list = [y for y in range(0, size_y)]
                self.line_list.clear()
                self.screen.clear()
                self.screen.refresh()
                continue
            self.add_lines(size_y, size_x)
            if self.color_mode == "cycle":
                if next(self.color_cycle_count) == self.color_cycle_delay:
                    color = list(CURSES_COLOR.keys())[next(self.color_cycle)]
                    setup_curses_colors(color,
                                        self.args.background,
                                        self.args.over_ride)
                    self.color_cycle_count = itertools.count(start=0, step=1)
            if self.dir == "old scrolling":
                self.display_old_scrolling()
            else:
                self.display_normal_scrolling()
            if self.args.wakeup:
                self.handle_wake_up()
            if self.args.run_timer and datetime.datetime.now() >= end_time:
                break
            time.sleep(DELAY_SPEED[self.args.delay])
            if self.handle_input():
                break
        self.screen.erase()
        self.screen.refresh()


    def handle_input(self) -> bool:
        """
        Returns True: Break. Quit the matrix
        Returns False: Continue running the matrix
        """
        ch = self.screen.getch()
        if ch == -1:
            return False
        elif self.args.screen_saver:
            return True
        elif ch in [81, 113]:  # q, Q
            return True
        elif self.args.disable_keys:
            return False
        if ch in WAKE_UP_KEYS:
            self.keys_pressed.append(ch)
            if self.keys_pressed == WAKE_UP_KEYS:
                wake_up_neo(self.screen, self.args.test_mode)
                while self.screen.getch() != -1:  # clears out the buffer
                    pass
                self.keys_pressed = []
                self.screen.bkgd(self.args.bg_char, curses.color_pair(1))
                return False
            elif len(self.keys_pressed) >= 4:
                self.keys_pressed = []
                return False
        # commands
        if ch == 98:  # b
            self.args.bold_on = True
            self.args.bold_all = False
        elif ch == 66:  # B
            self.args.bold_all = True
            self.args.bold_on = False
        elif ch in [78, 110]:  # n or N
            self.args.bold_on = False
            self.args.bold_all = False
        elif ch in [114, 116, 121, 117, 105, 111, 112, 91]:
            # r, t, y, u, i, o, p, [
            self.args.color = CURSES_CH_CODES_COLOR[ch]
            setup_curses_colors(self.args.color, self.args.background,
                                self.args.over_ride)
            self.color_mode = "normal"
        elif ch in [82, 84, 89, 85, 73, 79, 80, 123]:
            # R, T, Y, U, I, O, P, {
            self.args.lead_color = CURSES_CH_CODES_COLOR[ch]
            curses_lead_color(self.args.lead_color, self.args.background,
                              self.args.over_ride)
        elif ch in [18, 20, 25, 21, 9, 15, 16, 27]:
            # ctrl R, T, Y, U, I, O, P, [
            self.args.background = CURSES_CH_CODES_COLOR[ch]
            setup_curses_colors(self.args.color, self.args.background,
                                self.args.over_ride)
            curses_lead_color(self.args.lead_color, self.args.background,
                              self.args.over_ride)
            self.screen.bkgd(self.args.bg_char, curses.color_pair(1))
        elif ch == 97:  # a
            self.args.async_scroll = not self.args.async_scroll
        elif ch == 109:  # m
            if self.color_mode in ["random", "normal", "cycle"]:
                self.color_mode = "multiple"
                setup_curses_colors("random", self.args.background,
                                    self.args.over_ride)
            else:
                self.color_mode = "normal"
                setup_curses_colors("green", self.args.background,
                                    self.args.over_ride)
        elif ch == 77:  # M
            if self.color_mode in ["multiple", "normal", "cycle"]:
                self.color_mode = "random"
                setup_curses_colors("random", self.args.background,
                                    self.args.over_ride)
            else:
                self.color_mode = "normal"
                setup_curses_colors("green", self.args.background,
                                    self.args.over_ride)
        elif ch == 99:  # c
            if self.color_mode in ["random", "multiple", "normal"]:
                self.color_mode = "cycle"
            else:
                self.color_mode = "normal"
        elif ch == 108:  # l
            if self.dir == "right" or self.dir == "left":
                return False
            if self.spacer == 1:
                self.spacer = 2
                self.x_list = [x for x in range(0, curses.COLS, self.spacer)]
                self.line_list.clear()
                self.screen.clear()
                self.screen.refresh()
            else:
                spacer = 1
                self.x_list = [x for x in range(0, curses.COLS, spacer)]
        elif ch == 101:  # e
            self.args.zero_one = False
            if self.args.ext or self.args.ext_only:
                self.args.ext = False
            else:
                self.args.ext = True
            self.char_set = build_character_set2(self.args)
        elif ch == 69:  # E
            self.args.zero_one = False
            self.args.ext_only = not self.args.ext_only
            self.char_set = build_character_set2(self.args)
        elif ch == 122 and not self.args.zero_one:  # z
            self.args.zero_one = True
            self.char_set = build_character_set2(self.args)
        elif ch == 90 and self.args.zero_one:  # Z
            self.args.zero_one = False
            self.char_set = build_character_set2(self.args)
        elif ch == 23:  # ctrl-w
            self.args.wakeup = not self.args.wakeup
        elif ch == 118:  # v
            if self.dir == "down":
                self.dir = "up"
            elif self.dir == "up":
                self.dir = "down"
            else:
                self.dir = "up"
            self.x_list = [x for x in range(0, curses.COLS, self.spacer)]
            self.line_list.clear()
            self.screen.clear()
            self.screen.refresh()
        elif ch == 115:  # s
            self.dir = "down" if self.dir == "old scroll" else "old scroll"
            self.x_list = [x for x in range(0, curses.COLS, self.spacer)]
            self.screen.clear()
            self.screen.refresh()
            self.line_list.clear()
            time.sleep(0.2)
        elif ch == 261:  # right arrow
            if self.dir != "right":
                self.dir = "right"
                self.line_list.clear()
                self.screen.clear()
                self.screen.refresh()
                time.sleep(0.4)
                self.y_list = [y for y in range(1, curses.LINES)]
        elif ch == 260:  # left arrow
            if self.dir != "left":
                self.dir = "left"
                self.line_list.clear()
                self.screen.clear()
                self.screen.refresh()
                time.sleep(0.4)
                self.y_list = [y for y in range(1, curses.LINES)]
        elif ch == 259:  # up arrow
            if self.dir != "up":
                self.dir = "up"
                self.line_list.clear()
                self.screen.clear()
                self.screen.refresh()
                time.sleep(0.4)
                self.x_list = [x for x in range(0, curses.COLS, self.spacer)]
        elif ch == 258:  # down arrow
            if self.dir != "down":
                self.dir = "down"
                self.line_list.clear()
                self.screen.clear()
                self.screen.refresh()
                time.sleep(0.3)
                self.x_list = [x for x in range(0, curses.COLS, self.spacer)]
        elif ch in [100, 68]:  # d, D
            self.args.zero_one = False
            self.args.bold_on = False
            self.args.bold_all = False
            self.args.background = "black"
            self.args.color = "green"
            self.args.lead_color = "white"
            self.args.ext = False
            self.args.ext_only = False
            setup_curses_colors(self.args.color, self.args.background,
                                self.args.over_ride)
            curses_lead_color(self.args.lead_color, self.args.background,
                              self.args.over_ride)
            self.color_mode = "normal"
            self.args.async_scroll = False
            self.args.delay = 4
            self.args.Katakana_only = False
            self.args.katakana = False
            if self.dir != "down":
                self.dir = "down"
                self.x_list = [x for x in range(0, curses.COLS, self.spacer)]
                self.screen.clear()
                self.screen.refresh()
                self.line_list.clear()
                time.sleep(0.2)
            self.args.do_not_clear = False
            self.args.italic = False
            if self.spacer == 2:
                self.spacer = 1
                self.x_list = [x for x in range(0, curses.COLS, self.spacer)]
            self.char_set = build_character_set2(self.args)
            self.args.bg_char = DEFAULT_BG_CHAR
            self.screen.bkgd(self.args.bg_char, curses.color_pair(1))
        elif (self.color_mode == "cycle" and
              ch in CURSES_CH_CODES_CYCLE_DELAY.keys()):
            self.color_cycle_delay = 100 * CURSES_CH_CODES_CYCLE_DELAY[ch]
        elif 48 <= ch <= 57:  # number keys 0 to 9
            self.args.delay = int(chr(ch))
        elif ch == 87:  # W
            self.args.do_not_clear = not self.args.do_not_clear
        elif ch == 119:  # w
            self.screen.clear()
            self.screen.refresh()
            self.line_list.clear()
            time.sleep(2)
            return False
        elif ch == 106:  # j
            self.args.italic = not self.args.italic
        elif ch == 102:  # f
            # Freeze the Matrix
            while True:
                ch = self.screen.getch()
                if ch == 102:
                    break
                elif ch in [81, 113]:  # q, Q
                    return True
        elif ch == 75:  # K
            self.args.zero_one = False
            self.args.Katakana_only = not self.args.Katakana_only
            self.char_set = build_character_set2(self.args)
        elif ch == 107:  # k
            self.args.zero_one = False
            self.args.Katakana_only = False
            self.args.katakana = not self.args.katakana
            self.char_set = build_character_set2(self.args)
        elif ch == 4:  # ctrl-d
            self.args.disable_keys = True

    def handle_wake_up(self) -> None:
        if self.wake_up_time <= 0:
            wake_up_neo(self.screen, self.args.test_mode)
            self.wake_up_time = random.randint(2000, 3000)
            while self.screen.getch() != -1:  # clears out the buffer
                ...
            self.screen.bkgd(self.args.bg_char, curses.color_pair(1))
        else:
            self.wake_up_time -= 1

    def add_lines(self, size_y: int, size_x: int) -> None:
        if self.dir == "right" or self.dir == "left":
            y = random.choice(self.y_list)
            self.line_list.append(SingleLine(y, 0, size_x, size_y,self.dir))
        elif self.dir == "old scrolling":
            if len(self.line_list) < size_x - 1 and len(self.x_list) > 3:
                for _ in range(2):
                    x = random.choice(self.x_list)
                    self.x_list.pop(self.x_list.index(x))
                    self.line_list.append(OldScrollingLine(x, size_x, size_y))
        else:  # down and up
            if len(self.line_list) < size_x - 1 and len(self.x_list) > 3:
                for _ in range(2):
                    x = random.choice(self.x_list)
                    self.x_list.pop(self.x_list.index(x))
                    self.line_list.append(
                        SingleLine(0, x, size_x, size_y, self.dir))

    def display_old_scrolling(self) -> None:
        remove_list = []
        for line in self.line_list:
            if self.args.bold_all:
                bold = curses.A_BOLD
            elif self.args.bold_on and line.bold:
                bold = curses.A_BOLD
            else:
                bold = curses.A_NORMAL

            italic = curses.A_ITALIC if self.args.italic else curses.A_NORMAL
            color = curses.color_pair(line.line_color_number)
            if lead := line.get_lead():
                self.screen.addstr(lead[0], lead[1], lead[2],
                              curses.color_pair(10) + bold + italic)
            if remove := line.delete_last():
                self.screen.addstr(remove[0], remove[1], self.args.bg_char)
                if line.x not in self.x_list:
                    self.x_list.append(line.x)
            location_char_list = line.get_next()
            for cell in location_char_list:
                self.screen.addstr(*cell, color + bold + italic)
            if line.okay_to_delete():
                remove_list.append(line)
        self. screen.refresh()
        for rem in remove_list:
            self.line_list.pop(self.line_list.index(rem))

    def display_normal_scrolling(self) -> None:
        remove_list = []
        for line in self.line_list:
            if self.args.async_scroll and not line.async_scroll_turn():
                # Not the line's turn in async scroll mode then
                # continue to the next line.
                continue
            if remove_line := line.delete_last():
                if self.args.do_not_clear is False:
                    self.screen.addstr(remove_line[0], remove_line[1], self.args.bg_char)
                if line.x not in self.x_list:
                    self.x_list.append(line.x)

            if self.args.bold_all:
                bold = curses.A_BOLD
            elif self.args.bold_on:
                if random.randint(1, 3) <= 1:
                    bold = curses.A_BOLD
                else:
                    bold = curses.A_NORMAL
            else:
                bold = curses.A_NORMAL

            italic = curses.A_ITALIC if self.args.italic else curses.A_NORMAL
            if self.color_mode == "random":
                color = curses.color_pair(random.randint(1, 7))
            else:
                color = curses.color_pair(line.line_color_number)
            if new_char := line.get_next():
                self.screen.addstr(new_char[0],
                              new_char[1],
                              random.choice(self.char_set),
                              color + bold + italic)
            if lead_char := line.get_lead():
                self.screen.addstr(lead_char[0], lead_char[1],
                              random.choice(self.char_set),
                              curses.color_pair(10) + bold + italic)
            if line.okay_to_delete():
                remove_list.append(line)
        self. screen.refresh()
        for rem in remove_list:
            self.line_list.pop(self.line_list.index(rem))

    @classmethod
    def check_screen_size(cls, size_y: int, size_x: int) -> None:
        if size_y < MIN_SCREEN_SIZE_Y:
            raise PyMatrixError("Error screen height is to short.")
        if size_x < MIN_SCREEN_SIZE_X:
            raise PyMatrixError("Error screen width is to narrow.")

    def setup_colors(self) -> None:
        setup_curses_wake_up_colors(self.args.over_ride)
        curses_lead_color(self.args.lead_color,
                          self.args.background, self.args.over_ride)
        if self.args.color_number is not None:
            setup_curses_color_number(self.args.color_number,
                                      self.args.background, self.args.over_ride)
        else:
            setup_curses_colors(self.args.color, self.args.background,
                                self.args.over_ride)


def build_character_set2(args: argparse.Namespace):
    if args.zero_one:
        new_list = ["0", "1"]
    elif args.ext_only and args.test_mode:
        new_list = ["Ä"]
    elif args.ext_only:
        new_list = EXT_CHAR_LIST
    elif args.Katakana_only and args.test_mode:
        new_list = ["ﾎ", "0"]
    elif args.Katakana_only:
        new_list = KATAKANA_CHAR_LIST + KATAKANA_CHAR_LIST_ADDON
    elif args.katakana and args.ext and args.test_mode:
        new_list = ["T", "ﾎ", "Ä"]
    elif args.katakana and args.ext:
        new_list = KATAKANA_CHAR_LIST + EXT_CHAR_LIST + CHAR_LIST
    elif args.ext and args.test_mode:
        new_list = ["Ä", "T"]
    elif args.ext:
        new_list = CHAR_LIST + EXT_CHAR_LIST
    elif args.katakana and args.test_mode:
        new_list = ["T", "ﾎ"]
    elif args.katakana:
        new_list = CHAR_LIST + KATAKANA_CHAR_LIST
    elif args.test_mode:
        new_list = ["T"]
    else:
        new_list = CHAR_LIST
    OldScrollingLine.update_char_list(new_list)
    return new_list


def curses_lead_color(color: str, bg_color: str, over_ride: bool) -> None:
    if over_ride:
        curses.init_pair(10, CURSES_OVER_RIDE_COLORS[color],
                         CURSES_OVER_RIDE_COLORS[bg_color])
    else:
        curses.init_pair(10, CURSES_COLOR[color], CURSES_COLOR[bg_color])


def setup_curses_color_number(
        color_num: int,
        bg_color: str,
        override: bool) -> None:

    if override:
        bg = CURSES_OVER_RIDE_COLORS[bg_color]
    else:
        bg = CURSES_COLOR[bg_color]

    color_list = [color_num for _ in range(7)]
    for x, c in enumerate(color_list):
        curses.init_pair(x + 1, c, bg)


def setup_curses_colors(color: str, bg_color: str, over_ride: bool) -> None:
    """ Init colors pairs in the curses. """
    if over_ride:
        curses_colors = CURSES_OVER_RIDE_COLORS
    else:
        curses_colors = CURSES_COLOR
    if color == "random":
        color_list = list(curses_colors.keys())
    else:
        color_list = [color for _ in range(7)]
    for x, c in enumerate(color_list):
        curses.init_pair(x + 1, curses_colors[c], curses_colors[bg_color])


def setup_curses_wake_up_colors(override: bool) -> None:
    if override:
        curses.init_pair(WAKE_UP_PAIR,
                         CURSES_OVER_RIDE_COLORS["green"],
                         CURSES_OVER_RIDE_COLORS["black"])
    else:
        curses.init_pair(WAKE_UP_PAIR,
                         CURSES_COLOR["green"],
                         CURSES_COLOR["black"])


def wake_up_neo(screen, test_mode: bool) -> None:
    z = 0.06 if test_mode else 1  # For test mode - shorter test time
    screen.erase()
    screen.bkgd(" ", curses.color_pair(WAKE_UP_PAIR))
    screen.refresh()
    time.sleep(3 * z)
    display_text(screen, "Wake up, Neo...", 0.08 * z, 7.0 * z)
    display_text(screen, "The Matrix has you...", 0.25 * z, 7.0 * z)
    display_text(screen, "Follow the white rabbit.", 0.1 * z, 7.0 * z)
    display_text(screen, "Knock, knock, Neo.", 0.01 * z, 3.0 * z)
    time.sleep(2 * z)


def display_text(screen, text: str, type_time: float, hold_time: float) -> None:
    for i, letter in enumerate(text, start=1):
        screen.addstr(1, i, letter,
                      curses.color_pair(WAKE_UP_PAIR) + curses.A_BOLD)
        screen.refresh()
        time.sleep(type_time)
    time.sleep(hold_time)
    screen.erase()
    screen.refresh()


def positive_int_zero_to_nine(value: str) -> int:
    """
    Used with argparse.
    Checks to see if value is positive int between 0 and 10.
    """
    msg = f"{value} is an invalid positive int value 0 to 9"
    try:
        int_value = int(value)
        if int_value < 0 or int_value >= 10:
            raise argparse.ArgumentTypeError(msg)
        return int_value
    except ValueError:
        raise argparse.ArgumentTypeError(msg)


def color_type(value: str) -> str:
    """
    Used with argparse
    Checks to see if the value is a valid color and returns
    the lower case color name.
    """
    lower_value = value.lower()
    if lower_value in CURSES_COLOR.keys():
        return lower_value
    raise argparse.ArgumentTypeError(f"{value} is an invalid color name")


def positive_int(value: str) -> int:
    """
    Used by argparse.
    Checks to see if the value is positive.
    """
    msg = f"{value} is an invalid positive int value"
    try:
        int_value = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError(msg)
    else:
        if int_value <= 0:
            raise argparse.ArgumentTypeError(msg)
    return int_value


def int_between_1_and_255(value: str) -> int:
    """
    Used by argparse. Checks to see if the value is between 1 and 255
    """
    msg = f"{value} is an invalid positive int between 1 and 255"
    try:
        int_value = int(value)
        if int_value < 1 or int_value > 255:
            raise argparse.ArgumentTypeError(msg)
        return int_value
    except ValueError:
        raise argparse.ArgumentTypeError(msg)


def background_character(value: str) -> str:
    """
    Used by argparse. Checks to see if only one character was entered
    """
    msg = f"{value} is invalid. Please enter only one character"
    if len(value) == 1:
        return value
    else:
        raise argparse.ArgumentTypeError(msg)


def list_colors() -> None:
    color_dict = {"red": "\033[91m", "green": "\033[92m", "blue": "\033[94m", "cyan": "\033[96m",
                  "yellow": "\033[93m", "magenta": "\033[95m", "white": "\033[97m", "black": "\033[90m"}
    colors = list(CURSES_COLOR.keys())
    for c in colors:
        print(f"{color_dict[c]}{c} ", end="")
    print("\033[0m")


def display_commands() -> None:
    print("Commands available during run")
    print("0 - 9  Delay time (0-Fast, 4-Default, 9-Slow)")
    print("q or Q To quit Pymatrix-rain")
    print("b      Bold characters on")
    print("B      Bold all characters")
    print("n      Bold off (Default)")
    print("a      Asynchronous like scrolling (normal scrolling only)")
    print("m      Multiple color mode")
    print("M      Multiple random color mode (normal scrolling only")
    print("c      Cycle colors")
    print("d      Restore all defaults")
    print("l      Toggle double space lines")
    print("e      Extended characters on and off")
    print("E      Extended characters only")
    print("k      Katakana characters like those from the movies if correct font is installed")
    print("K      Katakana characters only")
    print("z      1 and 0 Mode On")
    print("Z      1 and 0 Mode Off")
    print("f      Freeze the matrix (q will still quit")
    print("v      Toggle matrix scrolling up")
    print("W      Toggle do not clear screen (normal scrolling only)")
    print("w      Clear the screen, wait 2 seconds and restart")
    print("j      Toggle italic text")
    print("s      Toggle old school scrolling down only")
    print("ctrl-d Disable option keys. q will still quit The Matrix.")
    print("up arrow    Matrix scrolls down to up")
    print("down arrow  Matrix scrolls up to down (default)")
    print("right arrow Matrix scrolls from left to right")
    print("left arrow  Matrix scrolls from right to left")
    print("r,t,y,u,i,o,p,[   Set color")
    print("R,T,Y,U,I,O,P,{   Set lead character color")
    print("ctrl + r,t,y,u,i,o,p,[  Set background color")
    print("shift 0 - 9 Cycle color delay (0-Fast, 4-Default, 9-Slow)")


def argument_parsing(
        argv: Optional[Sequence[str]] = None) -> argparse.Namespace:
    """ Command line argument parsing. """
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", dest="delay",
                        type=positive_int_zero_to_nine,
                        default=4,
                        help="Set the delay (speed)"
                             " 0: Fast, 4: Default, 9: Slow")
    parser.add_argument("-b", dest="bold_on", action="store_true",
                        help="Bold characters on")
    parser.add_argument("-B", dest="bold_all", action="store_true",
                        help="All bold characters (overrides -b)")
    parser.add_argument("-s", dest="screen_saver", action="store_true",
                        help="Screen saver mode.  Any key will exit.")
    parser.add_argument("-a", dest="async_scroll", action="store_true",
                        help="enable asynchronous like scrolling "
                             "(normal scrolling only)")
    parser.add_argument("-S", dest="start_timer", type=positive_int, default=0,
                        metavar="SECONDS", help="Set start timer in seconds")
    parser.add_argument("-R", dest="run_timer", type=positive_int, default=0,
                        metavar="SECONDS", help="Set run timer in seconds")
    parser.add_argument("-C", dest="color", type=color_type, default="green",
                        help="Set color.  Default is green")
    parser.add_argument("-L", dest="lead_color",
                        type=color_type, default="white",
                        help="Set the lead character color.  Default is white")
    parser.add_argument("-m", dest="multiple_mode", action="store_true",
                        help="Multiple color mode")
    parser.add_argument("-M", dest="random_mode", action="store_true",
                        help="Multiple random color mode "
                             "(normal scrolling only")
    parser.add_argument("-c", dest="cycle", action="store_true",
                        help="cycle through the colors")
    parser.add_argument("-e", dest="ext", action="store_true",
                        help="use extended characters")
    parser.add_argument("-E", dest="ext_only", action="store_true",
                        help="use only extended characters (overrides -e)")
    parser.add_argument("-l", dest="double_space", action="store_true",
                        help="Double space lines")
    parser.add_argument("-z", dest="zero_one", action="store_true",
                        help="Show only zero and ones. Binary")
    parser.add_argument("--background", type=color_type, default="black",
                        help="set background color. Default is black.")
    parser.add_argument("--bg_char", type=background_character,
                        default=DEFAULT_BG_CHAR,
                        help="Enter one character to use as background."
                             " May need to use '' around the character")
    parser.add_argument("-v", "--reverse", action="store_true",
                        help="Reverse the matrix. "
                             "The matrix scrolls up (vertical)")
    parser.add_argument("--scroll_right", action="store_true",
                        help="Matrix scrolls from left to right")
    parser.add_argument("--scroll_left", action="store_true",
                        help="Matrix scrolls from right to left")
    parser.add_argument("-o", "--old_school_scrolling", action="store_true",
                        help="Old school scrolling. Scroll down only")
    parser.add_argument("-j", "--italic", action="store_true",
                        help="Italic characters")
    parser.add_argument("-k", "--katakana", action="store_true",
                        help="Use half width Katakana (as seen in the movie) "
                             "characters. NOTE: requires the correct"
                             " font to be installed to work")
    parser.add_argument("-K", "--Katakana_only", action="store_true",
                        help="Use only half width Katakana (as seen "
                             "in the movie) characters only. NOTE: requires"
                             " the correct font to be installed to work")
    parser.add_argument("-O", dest="over_ride", action="store_true",
                        help="Override terminal window colors by using color"
                             " numbers between 16 and 255. This requires 256"
                             " color support in the terminal to work.")
    parser.add_argument("-W", "--do_not_clear", action="store_true",
                        help="do not clear the screen (Normal scrolling only)")
    parser.add_argument("--color_number", type=int_between_1_and_255,
                        default=None,
                        metavar="number",
                        help="Enter a number between 1 and 255 to select"
                             " character color. Requires 256 color support in "
                             "the terminal to work. Changing colors or "
                             "background color will remove the color entered.")
    parser.add_argument("--disable_keys", action="store_true",
                        help="Disable keys except for Q to quit. Screensaver "
                             "mode will not be affected")
    parser.add_argument("--list_colors", action="store_true",
                        help="Show available colors and exit. ")
    parser.add_argument("--list_commands", action="store_true",
                        help="List Commands and exit")
    parser.add_argument("--version", action="version",
                        version=f"Version: {version}")

    parser.add_argument("--wakeup", action="store_true",
                        help=argparse.SUPPRESS)
    parser.add_argument("--test_mode", action="store_true",
                        help=argparse.SUPPRESS)
    return parser.parse_args(argv)


def main(argv: Optional[Sequence[str]] = None) -> None:
    args = argument_parsing(argv)

    if args.list_colors:
        list_colors()
        return
    if args.list_commands:
        display_commands()
        return

    time.sleep(args.start_timer)
    os.environ.setdefault('ESCDELAY', '25')  # 25 milliseconds
    try:
        curses.wrapper(Matrix, args)
    except KeyboardInterrupt:
        pass
    except PyMatrixError as e:
        print(e)
        return


if __name__ == "__main__":
    main()




import pytest

from pymatrix import pymatrix


@pytest.mark.parametrize("test_values, expected_results", [
    ([], 4), (["-d0"], 0), (["-d1"], 1), (["-d", "2"], 2), (["-d3"], 3),
    (["-d4"], 4), (["-d 5"], 5), (["-d6"], 6), (["-d7"], 7),
    (["-d8"], 8), (["-d9"], 9)
])
def test_argument_parsing_delay(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.delay == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], False), (["--test_mode"], True)
])
def test_argument_parsing_test_mode(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.test_mode == expected_results


# @pytest.mark.parametrize("test_values, expected_results", [
#     ([], False), (["--test_mode_ext"], True)
# ])
# def test_argument_parsing_test_mode_ext(test_values, expected_results):
#     result = pymatrix.argument_parsing(test_values)
#     assert result.test_mode_ext == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], False), (["-b"], True)
])
def test_argument_parsing_bold_on(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.bold_on == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], False), (["-B"], True)
])
def test_argument_parsing_bold_all(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.bold_all == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], False), (["-s"], True)
])
def test_argument_parsing_screen_save_mode(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.screen_saver == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    (["-Cred"], "red"), (["-C", "Green"], "green"), (["-C", "BLUE"], "blue"),
    (["-CyeLLOW"], "yellow"), (["-C", "magenta"], "magenta"),
    (["-CCyan"], "cyan"), (["-Cwhite"], "white"), ([], "green")
])
def test_argument_parsing_color(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.color == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    (["-Lred"], "red"), (["-L", "Green"], "green"), (["-L", "BLUE"], "blue"),
    (["-LyeLLOW"], "yellow"), (["-L", "magenta"], "magenta"),
    (["-LCyan"], "cyan"), (["-Lwhite"], "white"), ([], "white")
])
def test_argument_parsing_lead_color(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.lead_color == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], 0), (["-S1"], 1), (["-S5"], 5), (["-S", "20"], 20)
])
def test_argument_parsing_start_timer(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.start_timer == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], 0), (["-R1"], 1), (["-R5"], 5), (["-R", "20"], 20)
])
def test_argument_parsing_run_timer(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.run_timer == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], False), (["--list_colors"], True)
])
def test_argument_parsing_list_colors(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.list_colors == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], False), (["--list_commands"], True)
])
def test_argument_parsing_list_commands(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.list_commands == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], False), (["-a"], True)
])
def test_argument_parsing_async_scroll(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.async_scroll == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], False), (["-m"], True)
])
def test_argument_parsing_multiple_color_mode(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.multiple_mode == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], False), (["-M"], True)
])
def test_argument_parsing_multiple_random_color_mode(test_values,
                                                     expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.random_mode == expected_results


@pytest.mark.parametrize("test_values, expected_results", [
    ([], False), (["-c"], True)
])
def test_argument_parsing_cycle_through_colors(test_values, expected_results):
    result = pymatrix.argument_parsing(test_values)
    assert result.cycle == expected_results


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-e"], True)
])
def test_argument_parsing_extended_char(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.ext == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-E"], True)
])
def test_argument_parsing_extended_char(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.ext_only == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-l"], True),
])
def test_argument_parsing_double_space_lines(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.double_space == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["--wakeup"], True)
])
def test_argument_parsing_wakeup(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.wakeup == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-z"], True),
])
def test_argument_parsing_zero_one(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.zero_one == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["--disable_keys"], True)
])
def test_argument_parsing_disable_keys(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.disable_keys == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], "black",), (["--background", "blue"], "blue")
])
def test_argument_parsing_background_color(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.background == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-v"], True), (["--reverse"], True)
])
def test_argument_parsing_reverse(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.reverse == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-O"], True)
])
def test_argument_parsing_over_ride(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.over_ride == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-W"], True), (["--do_not_clear"], True)
])
def test_argument_parsing_do_not_clear(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.do_not_clear == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ("1", 1), ("40", 40), ("100", 100), ("156", 156), ("255", 255),
])
def test_argument_parsing_color_number(test_value, expected_result):
    result = pymatrix.argument_parsing(["--color_number", test_value])
    assert result.color_number == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-j"], True), (["--italic"], True)
])
def test_argument_parsing_italic(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.italic == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-K"], True), (["--Katakana_only"], True),
])
def test_argument_parsing_katakana_only(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.Katakana_only == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-k"], True), (["--katakana"], True),
])
def test_argument_parsing_katakana(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.katakana == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["--scroll_right"], True),
])
def test_argument_parsing_scroll_right(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.scroll_right == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["--scroll_left"], True),
])
def test_argument_parsing_scroll_right(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.scroll_left == expected_result


@pytest.mark.parametrize("test_value, expected_result", [
    ([], False), (["-o"], True), (["--old_school_scrolling"], True)
])
def test_argument_paring_old_school_scrolling(test_value, expected_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.old_school_scrolling == expected_result


@pytest.mark.parametrize("test_value, expect_result", [
    ([], " "), (["--bg_char", "-"], "-"), (["--bg_char", "+"], "+"),
    (["--bg_char", "*"], "*")
])
def test_argument_parsing_background_character(test_value, expect_result):
    result = pymatrix.argument_parsing(test_value)
    assert result.bg_char == expect_result


# testing helper functions
@pytest.mark.parametrize("test_values, expected_results", [
    ("0", 0), ("1", 1), ("2", 2), ("3", 3), ("4", 4),
    ("5", 5), ("6", 6), ("7", 7), ("8", 8), ("9", 9)
])
def test_positive_int_zero_to_nine_normal(test_values, expected_results):
    """ Tests that the delay time conversion formula is working. """
    result = pymatrix.positive_int_zero_to_nine(test_values)
    assert result == expected_results


@pytest.mark.parametrize("test_values", [
    "-5", "10", "100", "2.5", " ", "Test", "test&*#", "",
])
def test_positive_int_zero_to_nine_error(test_values):
    """ Testing delay_positive_int will raise an error. """
    with pytest.raises(pymatrix.argparse.ArgumentTypeError):
        pymatrix.positive_int_zero_to_nine(test_values)


@pytest.mark.parametrize("test_values, expected_results", [
    ("red", "red"), ("Green", "green"), ("BLUE", "blue"),
    ("yeLLOW", "yellow"), ("magenta", "magenta"),
    ("Cyan", "cyan"), ("whiTe", "white")
])
def test_color_type_normal(test_values, expected_results):
    result = pymatrix.color_type(test_values)
    assert result


@pytest.mark.parametrize("test_values", [
    "orange", "12", "who", "<>", "", " ", "ter8934", "834DFD"
])
def test_color_type_error(test_values):
    with pytest.raises(pymatrix.argparse.ArgumentTypeError):
        pymatrix.color_type(test_values)


@pytest.mark.parametrize("test_values, expected_results", [
    ("1", 1), ("2", 2), ("6", 6), ("20", 20), ("500", 500)
])
def test_positive_int_normal(test_values, expected_results):
    result = pymatrix.positive_int(test_values)
    assert result == expected_results


@pytest.mark.parametrize("test_values", [
    "0", "-3", "1.3", "0.4", "10.4", "a", "b", "", " ", "$", "time32"
])
def test_positive_int_error(test_values):
    with pytest.raises(pymatrix.argparse.ArgumentTypeError):
        pymatrix.positive_int(test_values)


@pytest.mark.parametrize("test_value, expected_result", [
    ("1", 1), ("30", 30), ("100", 100), ("167", 167), ("200", 200),
    ("250", 250), ("255", 255)
])
def test_int_between_1_and_255(test_value, expected_result):
    result = pymatrix.int_between_1_and_255(test_value)
    assert result == expected_result


@pytest.mark.parametrize("test_values", [
    "0", "256", "34.4", "Blue", "test", "-4", "1001", "", " ", "c40", "30c",
    "*", "-C", "&",
])
def test_int_between_1_and_255_error(test_values):
    with pytest.raises(pymatrix.argparse.ArgumentTypeError):
        pymatrix.int_between_1_and_255(test_values)


@pytest.mark.parametrize("value", [
    " ", "-", "+", "_", "?", "=", "X", "6", "["
])
def test_background_character(value):
    result = pymatrix.background_character(value)
    assert result == value


@pytest.mark.parametrize("test_value", [
    "--", "__", "xy", "11", "x1", " 1", "=-", "[]"
])
def test_background_character_error(test_value):
    with pytest.raises(pymatrix.argparse.ArgumentTypeError):
        pymatrix.background_character(test_value)




from unittest import mock

from pymatrix import pymatrix


def test_get_lead_down():
    line = pymatrix.SingleLine(0, 5, 20, 20, "down")
    lead_char = line.get_lead()
    assert lead_char is not None
    assert lead_char[0] == 0
    assert lead_char[1] == 5


def test_get_lead_down_off_screen():
    line = pymatrix.SingleLine(50, 5, 20, 20, "down")
    line.lead_y = 18
    line.get_lead()
    lead_char = line.get_lead()
    assert lead_char is None


def test_get_lead_down_increment():
    line = pymatrix.SingleLine(0, 5, 20, 20, "down")
    line.get_lead()
    assert line.lead_y == 1
    assert line.x == 5
    assert line.y == -1


def test_get_lead_up():
    line = pymatrix.SingleLine(0, 5, 20, 20, "up")
    lead_char = line.get_lead()
    assert lead_char is not None
    assert lead_char[0] == 18
    assert lead_char[1] == 5


def test_get_lead_up_off_screen():
    line = pymatrix.SingleLine(0, 5, 20, 20, "up")
    line.lead_y = 0
    line.get_lead()
    lead_char = line.get_lead()
    assert lead_char is None


def test_get_lead_up_increment():
    line = pymatrix.SingleLine(0, 5, 20, 20, "up")
    line.get_lead()
    assert line.lead_y == 17
    assert line.x == 5
    assert line.y == 19


def test_get_lead_right():
    line = pymatrix.SingleLine(5, 0, 20, 20, "right")
    lead_char = line.get_lead()
    assert lead_char is not None
    assert lead_char[0] == 5
    assert lead_char[1] == 0


def test_get_lead_right_off_screen():
    line = pymatrix.SingleLine(5, 0, 20, 20, "right")
    line.lead_x = 18
    line.get_lead()
    lead_char = line.get_lead()
    assert lead_char is None


def test_get_lead_right_increment():
    line = pymatrix.SingleLine(5, 0, 20, 20, "right")
    line.get_lead()
    assert line.lead_x == 1
    assert line.x == -1
    assert line.y == 5


def test_get_lead_left():
    line = pymatrix.SingleLine(5, 0, 20, 20, "left")
    lead_char = line.get_lead()
    assert lead_char is not None
    assert lead_char[0] == 5
    assert lead_char[1] == 18


def test_get_lead_left_off_screen():
    line = pymatrix.SingleLine(5, 0, 20, 20, "left")
    line.lead_x = 0
    line.get_lead()
    lead_char = line.get_lead()
    assert lead_char is None


def test_get_lead_left_increment():
    line = pymatrix.SingleLine(5, 0, 20, 20, "left")
    line.get_lead()
    assert line.lead_x == 17
    assert line.x == 19
    assert line.y == 5


def test_get_next_down():
    line = pymatrix.SingleLine(0, 5, 20, 20, "down")
    next_char = line.get_next()
    assert next_char is None


def test_get_next_down_second_loop():
    line = pymatrix.SingleLine(0, 5, 20, 20, "down")
    line.get_next()  # first loop
    next_char = line.get_next()
    assert next_char is not None
    assert next_char[0] == 0
    assert next_char[1] == 5


def test_get_next_down_off_screen():
    line = pymatrix.SingleLine(0, 5, 20, 20, "down")
    line.y = 18
    line.get_next()
    next_char = line.get_next()
    assert next_char is None


def test_get_next_down_increment():
    line = pymatrix.SingleLine(0, 5, 20, 20, "down")
    line.get_next()
    assert line.x == 5
    assert line.y == 0
    assert line.lead_y == 0


def test_get_next_right():
    line = pymatrix.SingleLine(5, 0, 20, 20, "right")
    next_char = line.get_next()
    assert next_char is None


def test_get_next_right_second_loop():
    line = pymatrix.SingleLine(5, 0, 20, 20, "right")
    line.get_next()  # first loop
    next_char = line.get_next()
    assert next_char is not None
    assert next_char[0] == 5
    assert next_char[1] == 0


def test_get_next_right_increment():
    line = pymatrix.SingleLine(5, 0, 20, 20, "right")
    line.get_next()
    assert line.x == 0
    assert line.y == 5
    assert line.lead_x == 0


def test_get_next_right_off_screen():
    line = pymatrix.SingleLine(0, 5, 20, 20, "right")
    line.x = 18
    line.get_next()
    next_char = line.get_next()
    assert next_char is None


def test_get_next_left():
    line = pymatrix.SingleLine(5, 0, 20, 20, "left")
    next_char = line.get_next()
    assert next_char is None


def test_get_next_left_second_loop():
    line = pymatrix.SingleLine(5, 0, 20, 20, "left")
    line.get_next()  # first loop
    next_char = line.get_next()
    assert next_char is not None
    assert next_char[0] == 5
    assert next_char[1] == 18


def test_get_next_left_increment():
    line = pymatrix.SingleLine(5, 0, 20, 20, "left")
    line.get_next()
    assert line.x == 18
    assert line.y == 5
    assert line.lead_x == 18


def test_get_next_left_off_screen():
    line = pymatrix.SingleLine(0, 5, 20, 20, "left")
    line.x = 0
    line.get_next()
    next_char = line.get_next()
    assert next_char is None


def test_get_next_up():
    line = pymatrix.SingleLine(0, 5, 20, 20, "up")
    next_char = line.get_next()
    assert next_char is None


def test_get_next_up_second_loop():
    line = pymatrix.SingleLine(0, 5, 20, 20, "up")
    line.get_next()  # first loop
    next_char = line.get_next()
    assert next_char is not None
    assert next_char[0] == 18
    assert next_char[1] == 5


def test_get_next_up_off_screen():
    line = pymatrix.SingleLine(0, 5, 20, 20, "up")
    line.y = 0
    line.get_next()
    next_char = line.get_next()
    assert next_char is None


def test_get_next_up_increment():
    line = pymatrix.SingleLine(0, 5, 20, 20, "up")
    line.get_next()
    assert line.x == 5
    assert line.y == 18
    assert line.lead_y == 18


def test_delete_last_down_none():
    line = pymatrix.SingleLine(0, 5, 20, 6, "down")
    delete_last = line.delete_last()
    assert delete_last is None


def test_delete_last_down_delete():
    line = pymatrix.SingleLine(0, 5, 20, 6, "down")
    line.last_y = 0  # patch
    delete_last = line.delete_last()
    assert delete_last is not None
    assert delete_last[0] == 0
    assert delete_last[1] == 5


def test_delete_last_down_stop():
    line = pymatrix.SingleLine(0, 5, 20, 6, "down")
    line.last_y = 4  # patch
    line.delete_last()
    delete_last = line.delete_last()
    assert delete_last is None


def test_delete_last_down_last_y_increment():
    line = pymatrix.SingleLine(0, 5, 20, 6, "down")
    line.delete_last()
    assert line.y == -1
    assert line.lead_y == 0
    assert line.last_y == -2


def test_delete_last_up_none():
    line = pymatrix.SingleLine(0, 5, 20, 6, "up")
    delete_last = line.delete_last()
    assert delete_last is None


def test_delete_last_up_delete():
    line = pymatrix.SingleLine(0, 5, 20, 6, "up")
    line.last_y = 4  # patch
    delete_last = line.delete_last()
    assert delete_last is not None
    assert delete_last[0] == 4
    assert delete_last[1] == 5


def test_delete_last_up_stop():
    line = pymatrix.SingleLine(0, 5, 20, 6, "up")
    line.last_y = 0  # patch
    line.delete_last()
    delete_last = line.delete_last()
    assert delete_last is None


def test_delete_last_up_last_y_increment():
    line = pymatrix.SingleLine(0, 5, 20, 6, "up")
    line.delete_last()
    assert line.y == 5
    assert line.lead_y == 4
    assert line.last_y == 5


def test_delete_last_right_none():
    line = pymatrix.SingleLine(5, 0, 20, 6, "right")
    delete_last = line.delete_last()
    assert delete_last is None


def test_delete_last_right_delete():
    line = pymatrix.SingleLine(5, 0, 20, 6, "right")
    line.last_x = 0  # patch
    delete_last = line.delete_last()
    assert delete_last is not None
    assert delete_last[0] == 5
    assert delete_last[1] == 0


def test_delete_last_right_stop():
    line = pymatrix.SingleLine(5, 0, 6, 20, "right")
    line.last_x = 4  # patch
    line.delete_last()
    delete_last = line.delete_last()
    assert delete_last is None


def test_delete_last_right_last_x_increment():
    line = pymatrix.SingleLine(5, 0, 6, 20, "right")
    line.delete_last()
    assert line.x == -1
    assert line.lead_x == 0
    assert line.last_x == -2


def test_delete_last_left_none():
    line = pymatrix.SingleLine(5, 0, 20, 6, "left")
    delete_last = line.delete_last()
    assert delete_last is None


def test_delete_last_left_delete():
    line = pymatrix.SingleLine(5, 0, 20, 6, "left")
    line.last_x = 18  # patch
    delete_last = line.delete_last()
    assert delete_last is not None
    assert delete_last[0] == 5
    assert delete_last[1] == 18


def test_delete_last_left_stop():
    line = pymatrix.SingleLine(5, 0, 6, 20, "left")
    line.last_x = 6  # patch
    line.delete_last()
    delete_last = line.delete_last()
    assert delete_last is None


def test_delete_last_left_last_x_increment():
    line = pymatrix.SingleLine(5, 0, 6, 20, "left")
    line.delete_last()
    assert line.x == 5
    assert line.lead_x == 4
    assert line.last_x == 6


def test_okay_to_delete_down_false():
    line = pymatrix.SingleLine(0, 5, 20, 6, "down")
    ok_delete = line.okay_to_delete()
    assert ok_delete is False


def test_okay_to_delete_down_true():
    line = pymatrix.SingleLine(0, 5, 20, 6, "down")
    line.last_y = 5  # patch
    line.delete_last()
    ok_delete = line.okay_to_delete()
    assert ok_delete is True


def test_okay_to_delete_up_false():
    line = pymatrix.SingleLine(0, 5, 20, 6, "up")
    ok_delete = line.okay_to_delete()
    assert ok_delete is False


def test_okay_to_delete_up_true():
    line = pymatrix.SingleLine(0, 5, 20, 6, "up")
    line.last_y = 0  # patch
    line.delete_last()
    ok_delete = line.okay_to_delete()
    assert ok_delete is True


def test_okay_to_delete_right_false():
    line = pymatrix.SingleLine(5, 0, 20, 6, "right")
    ok_delete = line.okay_to_delete()
    assert ok_delete is False


def test_okay_to_delete_right_true():
    line = pymatrix.SingleLine(5, 0, 6, 20, "right")
    line.last_x = 5  # patch
    line.delete_last()
    ok_delete = line.okay_to_delete()
    assert ok_delete is True


def test_okay_to_delete_left_false():
    line = pymatrix.SingleLine(5, 0, 20, 6, "left")
    ok_delete = line.okay_to_delete()
    assert ok_delete is False


def test_okay_to_delete_left_true():
    line = pymatrix.SingleLine(5, 0, 6, 20, "left")
    line.last_x = 0  # patch
    line.delete_last()
    ok_delete = line.okay_to_delete()
    assert ok_delete is True


def test_async_scroll_turn():
    line = pymatrix.SingleLine(0, 5, 20, 6, "down")
    line.async_scroll_rate = 1
    turn = line.async_scroll_turn()
    assert turn is False
    turn = line.async_scroll_turn()
    assert turn is True
    turn = line.async_scroll_turn()
    assert turn is False


def test_async_scroll_turn_right():
    line = pymatrix.SingleLine(0, 5, 20, 6, "right")
    line.async_scroll_rate = 1
    turn = line.async_scroll_turn()
    assert turn is False
    turn = line.async_scroll_turn()
    assert turn is True
    turn = line.async_scroll_turn()
    assert turn is False


def test_single_line_class_down_init():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(0, 5, 20, 6, "down")
        assert line.direction == "down"
        assert line.height == 4
        assert line.x == 5
        assert line.async_scroll_count == 0
        assert line.async_scroll_rate == 3
        assert line.line_color_number == 3
        assert line.lead_y == 0
        assert line.y == -1
        assert line.last_y == -3


def test_single_line_class_up_init():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(0, 5, 20, 6, "up")
        assert line.direction == "up"
        assert line.height == 4
        assert line.x == 5
        assert line.async_scroll_count == 0
        assert line.async_scroll_rate == 3
        assert line.line_color_number == 3
        assert line.lead_y == 4
        assert line.y == 5
        assert line.last_y == 6


def test_single_line_class_right_init():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(5, 0, 6, 20, "right")
        assert line.direction == "right"
        assert line.height == 18
        assert line.width == 5
        assert line.x == -1
        assert line.y == 5
        assert line.async_scroll_count == 0
        assert line.async_scroll_rate == 3
        assert line.line_color_number == 3
        assert line.lead_x == 0
        assert line.last_x == -3
        assert line.lead_y == 0
        assert line.last_y == 0


def test_single_line_class_left_init():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(5, 0, 6, 20, "left")
        assert line.direction == "left"
        assert line.height == 18
        assert line.width == 5
        assert line.x == 5
        assert line.y == 5
        assert line.async_scroll_count == 0
        assert line.async_scroll_rate == 3
        assert line.line_color_number == 3
        assert line.lead_y == 0
        assert line.last_y == 0
        assert line.lead_x == 4
        assert line.last_x == 7


def test_single_line_class_down_one_turn():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(0, 5, 20, 6, "down")
        lead_char = line.get_lead()
        assert lead_char == (0, 5)
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False


def test_single_line_class_up_one_turn():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(0, 5, 20, 6, "up")
        lead_char = line.get_lead()
        assert lead_char == (4, 5)
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False


def test_single_line_class_right_one_turn():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(5, 0, 6, 20, "right")
        lead_char = line.get_lead()
        assert lead_char == (5, 0)
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False


def test_single_line_class_left_one_turn():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(5, 0, 6, 20, "left")
        lead_char = line.get_lead()
        assert lead_char == (5, 4)
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False


def test_single_line_class_down_multiple():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(0, 5, 20, 6, "down")
        line.get_lead()
        line.get_next()
        line.delete_last()
        line.okay_to_delete()
        lead_char = line.get_lead()
        assert lead_char == (1, 5)
        next_char = line.get_next()
        assert next_char == (0, 5)
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (2, 5)
        next_char = line.get_next()
        assert next_char == (1, 5)
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (3, 5)
        next_char = line.get_next()
        assert next_char == (2, 5)
        delete_last = line.delete_last()
        assert delete_last == (0, 5)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (4, 5)
        next_char = line.get_next()
        assert next_char == (3, 5)
        delete_last = line.delete_last()
        assert delete_last == (1, 5)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char == (4, 5)
        delete_last = line.delete_last()
        assert delete_last == (2, 5)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last == (3, 5)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last == (4, 5)
        ok_delete = line.okay_to_delete()
        assert ok_delete is True


def test_single_line_class_up_multiple():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(0, 5, 20, 6, "up")
        line.get_lead()
        line.get_next()
        line.delete_last()
        line.okay_to_delete()
        lead_char = line.get_lead()
        assert lead_char == (3, 5)
        next_char = line.get_next()
        assert next_char == (4, 5)
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (2, 5)
        next_char = line.get_next()
        assert next_char == (3, 5)
        delete_last = line.delete_last()
        assert delete_last == (4, 5)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (1, 5)
        next_char = line.get_next()
        assert next_char == (2, 5)
        delete_last = line.delete_last()
        assert delete_last == (3, 5)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (0, 5)
        next_char = line.get_next()
        assert next_char == (1, 5)
        delete_last = line.delete_last()
        assert delete_last == (2, 5)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char == (0, 5)
        delete_last = line.delete_last()
        assert delete_last == (1, 5)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last == (0, 5)
        ok_delete = line.okay_to_delete()
        assert ok_delete is True


def test_single_line_class_right_multiple():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(5, 0, 6, 20, "right")
        line.get_lead()
        line.get_next()
        line.delete_last()
        line.okay_to_delete()
        lead_char = line.get_lead()
        assert lead_char == (5, 1)
        next_char = line.get_next()
        assert next_char == (5, 0)
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (5, 2)
        next_char = line.get_next()
        assert next_char == (5, 1)
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (5, 3)
        next_char = line.get_next()
        assert next_char == (5, 2)
        delete_last = line.delete_last()
        assert delete_last == (5, 0)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (5, 4)
        next_char = line.get_next()
        assert next_char == (5, 3)
        delete_last = line.delete_last()
        assert delete_last == (5, 1)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char == (5, 4)
        delete_last = line.delete_last()
        assert delete_last == (5, 2)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last == (5, 3)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last == (5, 4)
        ok_delete = line.okay_to_delete()
        assert ok_delete is True


def test_single_line_class_left_multiple():
    with mock.patch.object(pymatrix.random, "randint", return_value=3):
        line = pymatrix.SingleLine(5, 0, 6, 20, "left")
        line.get_lead()
        line.get_next()
        line.delete_last()
        line.okay_to_delete()
        lead_char = line.get_lead()
        assert lead_char == (5, 3)
        next_char = line.get_next()
        assert next_char == (5, 4)
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (5, 2)
        next_char = line.get_next()
        assert next_char == (5, 3)
        delete_last = line.delete_last()
        assert delete_last is None
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (5, 1)
        next_char = line.get_next()
        assert next_char == (5, 2)
        delete_last = line.delete_last()
        assert delete_last == (5, 4)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char == (5, 0)
        next_char = line.get_next()
        assert next_char == (5, 1)
        delete_last = line.delete_last()
        assert delete_last == (5, 3)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char == (5, 0)
        delete_last = line.delete_last()
        assert delete_last == (5, 2)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last == (5, 1)
        ok_delete = line.okay_to_delete()
        assert ok_delete is False

        lead_char = line.get_lead()
        assert lead_char is None
        next_char = line.get_next()
        assert next_char is None
        delete_last = line.delete_last()
        assert delete_last == (5, 0)
        ok_delete = line.okay_to_delete()
        assert ok_delete is True



from unittest import mock
import os
import pytest
import sys
from hecate import Runner
from time import sleep

from pymatrix import pymatrix


@pytest.fixture(autouse=True, scope='session')
def _pytest_readline_workaround():
    # Work around for side effect when pytest imports readline.
    # Side effect is setting environment variables columns and lines.
    assert 'readline' in sys.modules

    os.environ['COLUMNS'] = os.environ['LINES'] = ''
    del os.environ['COLUMNS'], os.environ['LINES']


def pymatrix_run(*args):
    options = [a for a in args]
    return ["python3", "pymatrix/pymatrix.py"] + options


def test_pymatrix_screen_test_mode():
    with Runner(*pymatrix_run("--test_mode", "-d1")) as h:
        h.await_text("T")
        sc = h.screenshot()
        assert "x" not in sc
        assert "Ä" not in sc
        assert "ﾎ" not in sc


def test_pymatrix_test_ext_only():
    with Runner(*pymatrix_run("--test_mode", "-E", "-d1")) as h:
        h.await_text("Ä")
        sc = h.screenshot()
        assert "x" not in sc
        assert "T" not in sc
        assert "ﾎ" not in sc


def test_pymatrix_test_ext_too():
    with Runner(*pymatrix_run("--test_mode", "-e", "-d1")) as h:
        h.await_text("Ä")
        h.await_text("T")
        sc = h.screenshot()
        assert "x" not in sc
        assert "ﾎ" not in sc


@pytest.mark.parametrize("test_key", ["Q", "q"])
def test_pymatrix_quit(test_key):
    with Runner(*pymatrix_run("--test_mode", "-d1")) as h:
        h.await_text("T")
        h.write(test_key)
        h.press("Enter")
        h.await_exit()


@pytest.mark.parametrize("test_key",
                         ["Q", "q", " ", "8", ":", "*", "M", "b", "I"])
def test_pymatrix_quit_screen_saver_mode(test_key):
    with Runner(*pymatrix_run("--test_mode", "-s", "-d1")) as h:
        h.default_timeout = 2
        h.await_text("T")
        h.write(test_key)
        h.press("Enter")
        h.await_exit()


@pytest.mark.parametrize("size", [50, 20, 13, 11, 10])
def test_pymatrix_screen_width_start(size):
    with Runner(*pymatrix_run("--test_mode"), width=size, height=50) as h:
        h.await_text("T")


@pytest.mark.parametrize("size", [9, 8, 7])
def test_pymatrix_screen_width_start_fail(size):
    with Runner("bash", width=size, height=50) as h:
        h.default_timeout = 4
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(". .venv/bin/activate")
        h.press("Enter")
        h.await_text("(.venv)")
        h.write("pymatrix-rain --test_mode -d1")
        h.press("Enter")
        h.await_text("Error screen width is to narrow.")


def test_pymatrix_screen_resize_adjust_width():
    with Runner(*pymatrix_run("--test_mode"), width=50, height=50) as h:
        h.await_text("T")
        h.tmux.execute_command('split-window', '-ht0', '-l', 25)
        h.await_text("T")
        h.tmux.execute_command('resize-pane', '-L', 10)
        h.await_text("T")
        h.tmux.execute_command('resize-pane', '-L', 5)
        h.await_text("Error screen width is to narrow.")
        sc = h.screenshot()
        assert "T" not in sc
        assert "Error" in sc


@pytest.mark.parametrize("size", [20, 12, 11, 10])
def test_pymatrix_screen_height_start(size):
    with Runner(*pymatrix_run("--test_mode"), width=50, height=size) as h:
        h.await_text("T")


@pytest.mark.parametrize("size", [9, 8, 5])
def test_pymatrix_screen_height_start_fail(size):
    with Runner("bash", width=50, height=size) as h:
        h.default_timeout = 4
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(". .venv/bin/activate")
        h.press("Enter")
        h.await_text("(.venv)")
        h.write("pymatrix-rain --test_mode -d1")
        h.press("Enter")
        h.await_text("Error screen height is to short.")


def test_pymatrix_screen_resize_adjust_height():
    with Runner(*pymatrix_run("--test_mode"), width=50, height=50) as h:
        sleep(0.1)
        h.default_timeout=5
        h.await_text("T")
        h.tmux.execute_command('split-window', '-vt0', '-l', 25)
        h.await_text("T")
        h.tmux.execute_command('resize-pane', '-U', 10)
        h.await_text("T")
        h.tmux.execute_command('resize-pane', '-U', 8)
        h.await_text("Error screen height is to short.")
        sc = h.screenshot()
        assert "T" not in sc
        assert "Error screen height is to short." in sc


def test_pymatrix_screen_resize_height_too_short():
    with Runner("bash", width=50, height=11) as h:
        h.default_timeout = 4
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(". .venv/bin/activate")
        h.press("Enter")
        h.await_text("(.venv)")
        h.write("pymatrix-rain --test_mode -d1")
        h.press("Enter")
        sleep(0.1)
        h.await_text("T")
        h.tmux.execute_command("resize-window", "-y", "5")
        h.await_text("Error screen height is to short.")
        sc = h.screenshot()
        assert "T" not in sc


def test_pymatrix_start_timer():
    with Runner(*pymatrix_run("--test_mode", "-S2")) as h:
        h.default_timeout = 3
        h.await_text("T")


def test_pymatrix_run_timer():
    with Runner(*pymatrix_run("--test_mode", "-R2")) as h:
        h.default_timeout = 3
        h.await_text("T")
        h.await_exit()


@pytest.mark.parametrize("test_key", ["Q", "q"])
def test_pymatrix_quit_with_run_timer(test_key):
    with Runner(*pymatrix_run("--test_mode", "-R3", "-d1")) as h:
        h.await_text("T")
        h.write(test_key)
        h.press("Enter")
        h.await_exit()


def test_pymatrix_list_colors():
    with Runner(*pymatrix_run("--list_colors")) as h:
        h.default_timeout = 4
        h.await_text("red green blue yellow magenta cyan white")


def test_pymatrix_list_commands():
    with Runner(*pymatrix_run("--list_commands"), width=80, height=80) as h:
        h.await_text("Commands available during run")


def test_pymatrix_version():
    with Runner(*pymatrix_run("--version")) as h:
        h.await_text(f"Version: {pymatrix.version}")


def test_pymatrix_help():
    with Runner(*pymatrix_run("--help"), width=80, height=80) as h:
        h.await_text("usage:")


def test_wakeup_help_suppressed():
    with Runner(*pymatrix_run("--help"), width=50, height=50) as h:
        h.await_text("--version")
        screen_shot = h.screenshot()
        assert "wakeup" not in screen_shot


def test_pymatrix_setup_curses_colors():
    with mock.patch.object(pymatrix.curses,
                           "init_pair", return_value=None) as mock_pair:
        pymatrix.setup_curses_colors("random", "black", False)
        assert mock_pair.call_count == 8


def test_curses_lead_color():
    with mock.patch.object(pymatrix.curses,
                           "init_pair", return_value=None) as mock_pair:
        pymatrix.curses_lead_color("blue", "black", False)
        assert mock_pair.call_count == 1


def test_pymatrix_setup_curses_colors_override():
    with mock.patch.object(pymatrix.curses,
                           "init_pair", return_value=None) as mock_pair:
        pymatrix.setup_curses_colors("random", "black", True)
        assert mock_pair.call_count == 8


def test_curses_lead_color_override():
    with mock.patch.object(pymatrix.curses,
                           "init_pair", return_value=None) as mock_pair:
        pymatrix.curses_lead_color("blue", "black", True)
        assert mock_pair.call_count == 1


def test_pymatrix_setup_color_number():
    with mock.patch.object(pymatrix.curses,
                           "init_pair", return_value=None) as mock_pair:
        pymatrix.setup_curses_color_number(50, "black", False)
        assert mock_pair.call_count == 7


def test_pymatrix_setup_color_number_override():
    with mock.patch.object(pymatrix.curses,
                           "init_pair", return_value=None) as mock_pair:
        pymatrix.setup_curses_color_number(50, "black", True)
        assert mock_pair.call_count == 7


def test_pymatrix_setup_color_number_color():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_color_number(60, "black", False)
    assert pymatrix.curses.pair_content(1) == (60, 0)
    assert pymatrix.curses.pair_content(7) == (60, 0)


def test_pymatrix_setup_color_number_color_override():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_color_number(60, "black", True)
    assert pymatrix.curses.pair_content(1) == (60, 16)
    assert pymatrix.curses.pair_content(7) == (60, 16)


def test_pymatrix_setup_curses_colors_call():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_colors("green", "black", False)
    assert pymatrix.curses.pair_content(1) == (2, 0)
    assert pymatrix.curses.pair_content(7) == (2, 0)


def test_pymatrix_setup_curses_colors_call_background():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_colors("green", "blue", False)
    assert pymatrix.curses.pair_content(1) == (2, 4)
    assert pymatrix.curses.pair_content(7) == (2, 4)


def test_pymatrix_setup_curses_colors_no_wake_up_set():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_colors("green", "black", False)
    assert pymatrix.curses.pair_content(21) == (0, 0)
    assert pymatrix.curses.pair_content(pymatrix.WAKE_UP_PAIR) == (0, 0)


def test_pymatrix_setup_curses_wake_up_colors():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_wake_up_colors(False)
    assert pymatrix.curses.pair_content(pymatrix.WAKE_UP_PAIR) == (2, 0)


def test_pymatrix_setup_curses_wake_up_colors_override():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_wake_up_colors(True)
    assert pymatrix.curses.pair_content(pymatrix.WAKE_UP_PAIR) == (40, 16)


def test_pymatrix_setup_curses_colors_random():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_colors("random", "black", False)
    assert pymatrix.curses.pair_content(1) == (1, 0)
    assert pymatrix.curses.pair_content(2) == (2, 0)
    assert pymatrix.curses.pair_content(7) == (7, 0)
    assert pymatrix.curses.pair_content(8) == (0, 0)


def test_pymatrix_setup_curses_colors_random_background():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_colors("random", "cyan", False)
    assert pymatrix.curses.pair_content(1) == (1, 6)
    assert pymatrix.curses.pair_content(2) == (2, 6)
    assert pymatrix.curses.pair_content(7) == (7, 6)
    assert pymatrix.curses.pair_content(8) == (0, 6)


def test_pymatrix_setup_curses_colors_random_override():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_colors("random", "black", True)
    assert pymatrix.curses.pair_content(1) == (160, 16)
    assert pymatrix.curses.pair_content(2) == (40, 16)
    assert pymatrix.curses.pair_content(7) == (255, 16)
    assert pymatrix.curses.pair_content(8) == (16, 16)


def test_pymatrix_setup_curses_colors_random_background_override():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.setup_curses_colors("random", "cyan", True)
    assert pymatrix.curses.pair_content(1) == (160, 44)
    assert pymatrix.curses.pair_content(2) == (40, 44)
    assert pymatrix.curses.pair_content(7) == (255, 44)
    assert pymatrix.curses.pair_content(8) == (16, 44)


def test_pymatrix_curses_lead_color():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.curses_lead_color("white", "black", False)
    assert pymatrix.curses.pair_content(10) == (7, 0)


def test_pymatrix_curses_lead_color_blue():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.curses_lead_color("blue", "black", False)
    assert pymatrix.curses.pair_content(10) == (4, 0)


def test_pymatrix_curses_lead_color_background():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.curses_lead_color("white", "cyan", False)
    assert pymatrix.curses.pair_content(10) == (7, 6)


def test_pymatrix_curses_lead_color_blue_background():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.curses_lead_color("blue", "white", False)
    assert pymatrix.curses.pair_content(10) == (4, 7)


def test_pymatrix_curses_lead_color_override():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.curses_lead_color("white", "black", True)
    assert pymatrix.curses.pair_content(10) == (255, 16)


def test_pymatrix_curses_lead_color_blue_override():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.curses_lead_color("blue", "black", True)
    assert pymatrix.curses.pair_content(10) == (21, 16)


def test_pymatrix_curses_lead_color_background_override():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.curses_lead_color("white", "cyan", True)
    assert pymatrix.curses.pair_content(10) == (255, 44)


def test_pymatrix_curses_lead_color_blue_background_override():
    pymatrix.curses.initscr()
    pymatrix.curses.start_color()
    pymatrix.curses_lead_color("blue", "white", True)
    assert pymatrix.curses.pair_content(10) == (21, 255)


def test_pymatrix_display_commands(capsys):
    pymatrix.display_commands()
    captured_output = capsys.readouterr().out
    expected_text = "Commands available during run"
    assert expected_text in captured_output


def test_pymatrix_display_command_quit(capsys):
    pymatrix.display_commands()
    captured_output = capsys.readouterr().out
    expected_text = "q or Q To quit Pymatrix-rain"
    assert expected_text in captured_output


def test_pymatrix_control_c_running():
    # this test can be flaky. Might not be the best test for this.
    with Runner("bash") as h:
        h.default_timeout = 4
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(". .venv/bin/activate")
        h.press("Enter")
        h.await_text("(.venv)")
        h.write("pymatrix-rain --test_mode -d1")
        h.press("Enter")
        sleep(0.1)
        h.await_text("T")
        h.press("C-c")
        captured = h.screenshot()
        assert "Traceback" not in captured


def test_pymatrix_wakeup():
    # this is a long test
    with Runner(*pymatrix_run("--test_mode", "--wakeup")) as h:
        h.await_text("T")
        h.default_timeout = 10
        h.await_text("Wake up, Neo...")
        h.await_text("The Matrix has you...")
        h.await_text("Follow the white rabbit.")
        h.await_text("Knock, knock, Neo.")
        h.await_text("T")


def test_pymatrix_wakeup_key_command():
    # this is a long test
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.await_text("T")
        h.press("C-w")
        h.default_timeout = 10
        h.await_text("Wake up, Neo...")
        h.await_text("The Matrix has you...")
        h.await_text("Follow the white rabbit.")
        h.await_text("Knock, knock, Neo.")
        h.await_text("T")


def test_pymatrix_wakeup_now_keys():
    # this is a long test
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.await_text("T")
        h.default_timeout = 10
        h.write("w")
        h.write("A")
        h.write("k")
        h.write("e")
        h.await_text("Wake up, Neo...")
        h.await_text("The Matrix has you...")
        h.await_text("Follow the white rabbit.")
        h.await_text("Knock, knock, Neo.")
        h.await_text("T")


def test_pymatrix_wakeup_now_keys_ext_char():
    # this is a long test
    with Runner(*pymatrix_run("--test_mode", "-e")) as h:
        h.await_text("T")
        h.await_text("Ä")
        h.default_timeout = 10
        h.write("w")
        h.write("A")
        h.write("k")
        h.write("e")
        h.await_text("Wake up, Neo...")
        h.await_text("The Matrix has you...")
        h.await_text("Follow the white rabbit.")
        h.await_text("Knock, knock, Neo.")
        h.await_text("T")
        h.await_text("Ä")


def test_pymatrix_wakeup_now_keys_katakana_only():
    # this is a long test
    with Runner(*pymatrix_run("--test_mode", "-K")) as h:
        h.await_text("ﾎ")
        h.await_text("0")
        h.default_timeout = 10
        h.write("w")
        h.write("A")
        h.write("k")
        h.write("e")
        h.await_text("Wake up, Neo...")
        h.await_text("The Matrix has you...")
        h.await_text("Follow the white rabbit.")
        h.await_text("Knock, knock, Neo.")
        h.await_text("ﾎ")
        h.await_text("T")


def test_pymatrix_wakeup_do_not_quit_on_q():
    # this is a long test
    with Runner(*pymatrix_run("--test_mode", "--wakeup")) as h:
        h.await_text("T")
        h.default_timeout = 10
        h.await_text("Wake up, Neo...")
        h.write("q")
        h.press("Enter")
        h.await_text("Knock, knock, Neo.")
        h.await_text("T")


def test_pymatrix_wakeup_ignore_multiple_key_presses():
    # this is a long test
    with Runner(*pymatrix_run("--test_mode", "--wakeup")) as h:
        h.await_text("T")
        h.default_timeout = 10
        h.await_text("Wake up, Neo...")
        h.write("q")
        h.await_text("The Matrix has you...")
        h.write("q")
        h.await_text("Knock, knock, Neo.")
        h.await_text("T")


def test_pymatrix_wakeup_ignore_alot_of_key_presses():
    # this is a long test
    with Runner(*pymatrix_run("--test_mode", "--wakeup")) as h:
        h.await_text("T")
        h.default_timeout = 10
        h.await_text("Wake up, Neo...")
        h.write("q")
        h.await_text("The Matrix has you...")
        h.write("q")
        h.await_text("Follow the white rabbit.")
        h.write("q")
        h.write("q")
        h.write("q")
        h.write("q")
        h.write("q")
        h.await_text("Knock, knock, Neo.")
        h.await_text("T")


def test_pymatrix_wakeup_now_ignore_keys_multiple_key_presses():
    # this is a long test
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.await_text("T")
        h.default_timeout = 10
        h.write("w")
        h.write("A")
        h.write("k")
        h.write("e")
        h.await_text("Wake up, Neo...")
        h.write("q")
        h.await_text("The Matrix has you...")
        h.write("q")
        h.await_text("Follow the white rabbit.")
        h.write("q")
        h.write("q")
        h.await_text("Knock, knock, Neo.")
        h.await_text("T")


@pytest.mark.parametrize("width", [80, 50, 40, 30, 20, 15, 10])
def test_pymatrix_double_space(width):
    with Runner(*pymatrix_run("--test_mode", "-l"),
                width=width, height=50) as h:
        h.await_text("T")
        sc = h.screenshot()
        assert "Traceback" not in sc


def test_pymatrix_zero_one_command_line_double_space():
    with Runner(*pymatrix_run("-z", "-l"), width=75, height=50) as h:
        h.await_text("0")
        h.await_text("1")
        sc = h.screenshot()
        assert "T" not in sc
        assert "5" not in sc


def test_pymatrix_zero_one_command_line():
    with Runner(*pymatrix_run("-z"), width=75, height=50) as h:
        h.await_text("0")
        h.await_text("1")
        sc = h.screenshot()
        assert "T" not in sc
        assert "5" not in sc


def test_pymatrix_zero_one_running_command():
    with Runner(*pymatrix_run("--test_mode", "-l"), width=100, height=25) as h:
        h.default_timeout = 4
        h.await_text("T")
        sleep(0.5)
        h.write("z")
        h.press("Enter")
        sleep(0.5)
        sc = h.screenshot()
        assert "T" in sc
        assert "0" in sc
        assert "1" in sc
        sleep(3)
        sc = h.screenshot()
        assert "T" not in sc
        sleep(0.5)
        h.write("Z")
        h.press("Enter")
        sleep(0.5)
        sc = h.screenshot()
        assert "T" in sc
        assert "0" in sc
        assert "1" in sc
        sleep(3)
        sc = h.screenshot()
        h.default_timeout = 6
        assert "T" in sc


def test_pymatrix_no_zero_one_running():
    with Runner(*pymatrix_run("--test_mode"), width=100, height=50) as h:
        h.await_text("T")
        h.write("Z")
        h.press("Enter")
        sleep(2)
        h.default_timeout = 3
        h.await_text("T")
        sc = h.screenshot()
        for letter in "AaBbCcDe0987654321ZzRrOoPpQqWweEY" \
                      "yUuIiOoPpKkLlJjmMnNXxSsgGhH":
            assert letter not in sc


def test_pymatrix_disable_keys():
    with Runner(*pymatrix_run("--test_mode", "--disable_keys")) as h:
        h.default_timeout = 2
        h.await_text("T")
        h.write("z")
        h.press("Enter")
        sleep(0.2)
        h.await_text("T")
        sc = h.screenshot()
        assert "1" not in sc


def test_pymatrix_disable_keys_quit():
    with Runner(*pymatrix_run("--test_mode", "--disable_keys")) as h:
        h.await_text("T")
        h.write("Q")
        h.await_exit()


def test_pymatrix_disable_keys_and_screen_saver():
    with Runner(*pymatrix_run("--test_mode", "-s", "--disable_keys")) as h:
        h.await_text("T")
        h.write("d")
        h.await_exit()


def test_pymatrix_disable_keys_check_running_keys():
    # not the best test
    # flaky test
    with Runner(*pymatrix_run("--test_mode", "--disable_keys")) as h:
        h.default_timeout = 5
        h.await_text("T")
        for k in "abcdefghijklmnoprstuvwxyzABCDEFGHIJKLMNOPRSTUVWXYZ" \
                 "1234567890{[}]!@#$%^&*()_+=-~` ":
            h.write(k)
            h.press("Enter")
            sleep(0.05)
            h.await_text("T")
        h.write("Q")
        h.press("Enter")
        h.await_exit()


def test_pymatrix_freeze():
    with Runner(*pymatrix_run("--test_mode"), width=100, height=10) as h:
        h.default_timeout = 2
        h.await_text("T")
        h.write("f")
        sleep(0.05)
        sc = h.screenshot()
        sleep(1)
        sc2 = h.screenshot()
        assert sc == sc2
        h.write("f")
        sleep(0.5)
        sc3 = h.screenshot()
        assert sc3 != sc2


def test_pymatrix_freeze_quit():
    with Runner(*pymatrix_run("--test_mode"), width=100, height=10) as h:
        h.default_timeout = 2
        h.await_text("T")
        h.write("f")
        sleep(0.05)
        h.write("q")
        h.await_exit()


def test_pymatrix_freeze_no_other_keys():
    with Runner(*pymatrix_run("--test_mode"), width=100, height=10) as h:
        h.default_timeout = 2
        h.await_text("T")
        h.write("f")
        sleep(0.05)
        sc = h.screenshot()
        h.write("z")
        sleep(0.5)
        assert h.screenshot() == sc


def test_pymatrix_freeze_scroll_right():
    with Runner(*pymatrix_run("--test_mode", "--scroll_right"),
                width=30, height=30) as h:
        h.default_timeout = 2
        h.await_text("T")
        h.write("f")
        sleep(0.05)
        sc = h.screenshot()
        sleep(1)
        assert h.screenshot() == sc


@pytest.mark.parametrize("test_value", ["-v", "--reverse"])
def test_pymatrix_reverse(test_value):
    with Runner(*pymatrix_run("--test_mode", test_value),
                width=100, height=10) as h:
        h.await_text("T")


def test_pymatrix_reverse_key():
    with Runner(*pymatrix_run("--test_mode"), width=100, height=10) as h:
        h.await_text("T")
        h.press("v")
        h.await_text("T")


def test_pymatrix_scroll_right_command_line():
    with Runner(*pymatrix_run("--test_mode", "--scroll_right"),
                width=20, height=30) as h:
        h.default_timeout = 3
        sleep(0.1)
        h.await_text("T")
        sleep(0.2)
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        column_left = []
        for line in lines:
            if len(line) == 0:
                continue
            column_left.append(line[0])
        assert "T" in column_left


def test_pymatrix_scroll_left_command_line():
    with Runner(*pymatrix_run("--test_mode", "--scroll_left"),
                width=50, height=30) as h:
        h.default_timeout = 3
        sleep(0.05)
        h.await_text("T")
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(f"{line:>50}")
        column_left = []
        column_right = []
        for line in lines:
            if len(line) == 0:
                continue
            column_left.append(line[0])
            column_right.append(line[-1])
        assert "T" in column_right
        assert "T" not in column_left


def test_pymatrix_change_direction_from_down_to_right():
    # flaky test
    with Runner(*pymatrix_run("--test_mode"), width=50, height=10) as h:
        h.default_timeout = 3
        h.await_text("T")
        sleep(0.1)
        h.press("Right")
        sleep(0.3)
        h.await_text("T")
        sleep(0.4)
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        column_left = []
        for line in lines:
            if len(line) == 0:
                continue
            column_left.append(line[0])
        assert "T" in column_left


def test_pymatrix_change_direction_from_down_to_left():
    with Runner(*pymatrix_run("--test_mode"), width=20, height=30) as h:
        h.default_timeout = 4
        h.await_text("T")
        sleep(0.3)
        h.press("Left")
        h.press("Enter")
        sleep(0.3)
        h.await_text("T")
        sleep(0.2)
        h.await_text("T")
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(f"{line:>50}")
        column_left = []
        column_right = []
        for line in lines:
            if len(line) == 0:
                continue
            column_left.append(line[0])
            column_right.append(line[-1])
        assert "T" in column_right
        assert "T" not in column_left


def test_pymartix_change_direction_from_right_to_down():
    with Runner(*pymatrix_run("--test_mode", "--scroll_right"),
                width=20, height=30) as h:
        h.default_timeout = 3
        h.await_text("T")
        sleep(0.1)
        h.press("Down")
        sleep(0.3)
        h.await_text("T")
        sleep(0.4)
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        assert "T" in lines[0] or "T" in lines[1]


def test_pymatrix_scroll_right_default_key():
    with Runner(*pymatrix_run("--test_mode", "--scroll_right"),
                width=20, height=30) as h:
        h.default_timeout = 3
        h.await_text("T")
        sleep(0.1)
        h.press("d")
        sleep(0.3)
        h.await_text("T")
        sleep(0.1)
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        assert "T" in lines[0] or "T" in lines[1]


def test_pymatrix_normal_char_at_top():
    with Runner(*pymatrix_run("--test_mode"), width=50, height=10) as h:
        h.await_text("T")
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        assert "T" in lines[0] or "T" in lines[1]
        assert "T" not in lines[-2] and "T" not in lines[-1]


@pytest.mark.parametrize("test_value", ["-v", "--reverse"])
def test_pymatrix_scrolling_up_chars_at_bottom(test_value):
    with Runner(*pymatrix_run("--test_mode", test_value),
                width=50, height=20) as h:
        h.await_text("T")
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        assert any(["T" in line for line in lines[12:]])
        assert "T" not in lines[0] and "T" not in lines[1]


def test_pymatrix_change_dir():
    # flaky test
    with Runner(*pymatrix_run("--test_mode", "-d9"), width=50, height=10) as h:
        h.default_timeout = 4
        h.await_text("T")
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        assert "T" in lines[0] or "T" in lines[1]
        assert "T" not in lines[-2] and "T" not in lines[-1]
        sleep(0.1)
        h.press("v")
        h.await_text("T")
        sleep(0.2)
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        assert "T" in lines[-2] or "T" in lines[-1]
        assert "T" not in lines[0] and "T" not in lines[1]


def test_pymatrix_change_dir_up_to_down():
    # flaky test
    with Runner(*pymatrix_run("--test_mode", "-d9", "-v"),
                width=50, height=10) as h:
        h.default_timeout = 3
        h.await_text("T")
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        assert any(["T" in line for line in lines[5:]])
        assert "T" not in lines[0] and "T" not in lines[1]
        sleep(0.1)
        h.press("v")
        h.await_text("T")
        sleep(0.2)
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        assert "T" in lines[0] or "T" in lines[1]
        assert "T" not in lines[-2] and "T" not in lines[-1]


def test_pymatrix_change_dir_up_to_down_default_key():
    # flakey test
    with Runner(*pymatrix_run("--test_mode", "-d9", "-v"),
                width=50, height=10) as h:
        h.default_timeout = 3
        h.await_text("T")
        sleep(0.1)
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        assert any(["T" in line for line in lines[5:]])
        assert "T" not in lines[0] and "T" not in lines[1]
        sleep(0.1)
        h.press("d")
        h.await_text("T")
        sleep(0.4)
        sc = h.screenshot()
        lines = []
        for line in sc.splitlines():
            lines.append(line)
        assert "T" in lines[0] or "T" in lines[1]
        assert "T" not in lines[-2] and "T" not in lines[-1]


def test_pymatrix_clear_screen():
    with Runner(*pymatrix_run("--test_mode"), width=80, height=20) as h:
        h.default_timeout = 3
        h.await_text("T")
        h.press("w")
        h.press("Enter")
        sleep(0.5)
        sc = h.screenshot()
        assert "T" not in sc
        h.await_text("T")


def test_pymatrix_clear_screen_katakana():
    with Runner(*pymatrix_run("--test_mode", "-K"), width=80, height=20) as h:
        h.default_timeout = 3
        h.await_text("ﾎ")
        h.await_text("0")
        h.press("w")
        h.press("Enter")
        sleep(0.5)
        sc = h.screenshot()
        assert "T" not in sc
        assert "ﾎ" not in sc
        assert "0" not in sc
        h.await_text("ﾎ")
        h.await_text("0")


def test_pymatrix_restore_defaults_katakana():
    with Runner(*pymatrix_run("--test_mode", "-K")) as h:
        h.default_timeout = 4
        h.await_text("ﾎ")
        h.await_text("0")
        h.press("d")
        h.await_text("T")
        sleep(4)
        sc = h.screenshot()
        assert "ﾎ" not in sc


def test_pymatrix_restore_defaults_delay():
    with Runner(*pymatrix_run("--test_mode", "-d8")) as h:
        h.default_timeout = 1
        h.await_text("T")
        h.press("d")
        sleep(1)
        h.await_text("T")


@pytest.mark.parametrize("test_value", ["1", "100", "40", "255", "128"])
def test_pymatrix_valid_color_number(test_value):
    cmd = f"TERM=xterm-256color python3 pymatrix/pymatrix.py --test_mode" \
          f" --color_number {test_value}"
    with Runner("bash") as h:
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(cmd)
        h.press("Enter")
        h.await_text("T")


@pytest.mark.parametrize("test_value", ["0", "A", "blue", "256"])
def test_pymatrix_invalid_color_number(test_value):
    cmd = f"TERM=xterm-256color python3 pymatrix/pymatrix.py --test_mode" \
          f" --color_number {test_value}"
    with Runner("bash") as h:
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(cmd)
        h.press("Enter")
        h.await_text(f"{test_value} is an invalid positive int "
                     f"between 1 and 255")


def test_pymatrix_command_line_italic():
    cmd = f"TERM=xterm-256color python3 pymatrix/pymatrix.py --test_mode -j"
    with Runner("bash") as h:
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(cmd)
        h.press("Enter")
        h.await_text("T")


def test_pymatrix_command_line_italic_reverse():
    cmd = f"TERM=xterm-256color python3 pymatrix/pymatrix.py --test_mode -j -v"
    with Runner("bash") as h:
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(cmd)
        h.press("Enter")
        h.await_text("T")


def test_pymatrix_command_line_italic_zero_one():
    cmd = f"TERM=xterm-256color python3 pymatrix/pymatrix.py -z -j -v"
    with Runner("bash") as h:
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(cmd)
        h.press("Enter")
        h.await_text("0")
        h.await_text("1")


def test_pymatrix_default():
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.await_text("T")
        h.press("q")
        h.await_exit()


@pytest.mark.parametrize("args", [
    pytest.param(("-d", "9"), id="delay slow"),
    pytest.param(("-d", "1"), id="delay fast"),
    pytest.param(("-b", ), id="bold on"),
    pytest.param(("-B", ), id="bold all"),
    pytest.param(("-C", "blue"), id="color blue"),
    pytest.param(("-C", "yellow"), id="color yellow"),
    pytest.param(("-L", "red"), id="lead color red"),
    pytest.param(("-m", ), id="Multiple colors"),
    pytest.param(("-M", ), id="Multiple random colors"),
    pytest.param(("-c", ), id="cycle colors"),
    pytest.param(("--background", "white"), id="background color white"),
    pytest.param(("--reverse", ), id="reverse"),
    pytest.param(("-v", ), id="reverse -v"),
    pytest.param(("-W", ), id="do not clear -W"),
    pytest.param(("--do_not_clear", ), id="do not clear"),
    pytest.param(("--disable_keys", ), id="disable keys"),
    pytest.param(("-C", "black", "-L", "green", "--background", "white"),
                 id="color, lead and background"),
    pytest.param(("-C", "red", "-L", "yellow", "--background", "cyan", "-B"),
                 id="color, lead, background and bold all"),
    pytest.param(("--background", "white", "-M"),
                 id="background white and multiple random colors"),
])
def test_pymatrix_command_line_options(args):
    arguments = ("--test_mode", *args)
    with Runner(*pymatrix_run(*arguments)) as h:
        h.await_text("T")
        h.press("q")
        h.await_exit()


def test_pymatrix_test_katakana_ony():
    with Runner(*pymatrix_run("--test_mode", "-K")) as h:
        h.await_text("ﾎ")
        h.await_text("0")
        sc = h.screenshot()
        assert "T" not in sc
        assert "Ä" not in sc
        assert "5" not in sc
        assert "r" not in sc


def test_pymatrix_test_katakana_char():
    with Runner(*pymatrix_run("--test_mode", "-k")) as h:
        h.await_text("ﾎ")
        h.await_text("T")
        sc = h.screenshot()
        assert "Ä" not in sc
        assert "5" not in sc
        assert "r" not in sc
        assert "0" not in sc


def test_pymatrix_test_katakana_char_ext():
    with Runner(*pymatrix_run("--test_mode", "-k", "-e")) as h:
        h.await_text("ﾎ")
        h.await_text("T")
        h.await_text("Ä")
        sc = h.screenshot()
        assert "5" not in sc
        assert "r" not in sc
        assert "0" not in sc


def test_pymatrix_running_command_katakana_only():
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.await_text("T")
        h.press("K")
        h.await_text("ﾎ")
        h.await_text("0")


def test_pymatrix_running_command_katakana_only_switching_back():
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.default_timeout = 5
        h.await_text("T")
        h.press("K")
        h.await_text("ﾎ")
        h.await_text("0")
        sleep(4)
        h.press("k")
        h.await_text("T")


def test_pymatrix_running_command_katakana():
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.default_timeout = 3
        h.await_text("T")
        h.press("k")
        h.await_text("ﾎ")
        sleep(2)
        h.await_text("ﾎ")
        h.await_text("T")


def test_pymatrix_katakana_quit():
    with Runner(*pymatrix_run("--test_mode", "-K")) as h:
        h.await_text("ﾎ")
        h.press("q")
        h.await_exit()


def test_pymatrix_test_katakana_ony_double_columns():
    # double space lines
    with Runner(*pymatrix_run("--test_mode", "-K", "-l")) as h:
        h.await_text("ﾎ")
        h.await_text("0")


def test_pymatrix_katakana_italic():
    with Runner(*pymatrix_run("--test_mode", "-K", "-j")) as h:
        h.await_text("ﾎ")
        h.await_text("0")


def test_pymatrix_katakana_bold_all():
    with Runner(*pymatrix_run("--test_mode", "-K", "-B")) as h:
        h.await_text("ﾎ")
        h.await_text("0")


def test_pymatrix_katakana_reverse():
    with Runner(*pymatrix_run("--test_mode", "-K", "-v")) as h:
        h.await_text("ﾎ")
        h.await_text("0")


def test_delay_keys():
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.default_timeout = 3
        h.await_text("T")
        for k in "0123456789":
            sleep(0.5)
            sc = h.screenshot()
            h.press(k)
            h.await_text("T")
            sleep(0.1)
            assert sc != h.screenshot()


def test_pymatrix_old_scroll_test_mode():
    with Runner(*pymatrix_run("-o", "--test_mode")) as h:
        h.default_timeout = 3
        h.await_text("T")
        sleep(0.1)
        sc = h.screenshot()
        for k in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSUVWXYZ1234567890+=-&^$#*":
            assert k not in sc

def test_pymatrix_old_scroll_test_mode_extended_characters():
    with Runner(*pymatrix_run("-o", "-e", "--test_mode")) as h:
        h.default_timeout = 3
        h.await_text("T")
        h.await_text("Ä")


def test_pymatrix_old_scroll_test_mode_extended_characters_only():
    with Runner(*pymatrix_run("-o", "-E", "--test_mode")) as h:
        h.default_timeout = 3
        h.await_text("Ä")
        sleep(0.1)
        sc = h.screenshot()
        assert "T" not in sc


def test_pymatrix_old_scroll_key():
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.default_timeout = 2
        h.await_text("T")
        h.press("s")
        h.await_text("T")


def test_pymatrix_escape_key_delay():
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.default_timeout = 3
        h.await_text("T")
        h.press("Escape")
        sc = h.screenshot()
        sleep(1)
        assert h.screenshot() != sc


def test_pymatrix_disable_key_ctrl_d():
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.await_text("T")
        h.press("^d")
        h.await_text("T")
        h.press("K")
        sleep(0.5)
        h.await_text("T")
        sc = h.screenshot()
        assert "ﾎ" not in sc
        assert "T" in sc


def test_pymatrix_disable_key_ctrl_d_quit():
    with Runner(*pymatrix_run("--test_mode")) as h:
        h.await_text("T")
        h.press("^d")
        h.await_text("T")
        h.press("q")
        h.await_exit()


def test_pymatrix_background_character_default_character():
    with Runner(*pymatrix_run("--test_mode",)) as h:
        h.await_text("T")
        h.await_text(" ")


@pytest.mark.parametrize("test_value", [
    "_", "+", "*", "@", "$", ".", "-"
])
def test_pymatrix_background_character(test_value):
    with Runner(*pymatrix_run("--bg_char", test_value,
                              "--test_mode")) as h:
        h.await_text("T")
        h.await_text(test_value)
        assert " " not in h.screenshot()


@pytest.mark.parametrize("test_value", [
    "_", "+", "*", "@", "$", ".", "-"
])
def test_pymatrix_background_character_old_scrolling(test_value):
    with Runner(*pymatrix_run("--bg_char", test_value,
                              "--test_mode", "-o")) as h:
        h.await_text("T")
        h.await_text(test_value)
        assert " " not in h.screenshot()


@pytest.mark.parametrize("test_value", [
    "*", "|", "!", "<", "#", ">", "$", "^"
])
def test_pymatrix_background_character_quotes(test_value):
    with Runner("bash") as h:
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(". .venv/bin/activate")
        h.press("Enter")
        h.await_text("(.venv)")
        h.write(f"pymatrix-rain --test_mode --bg_char '{test_value}'")
        h.press("Enter")
        h.await_text("T")
        h.await_text(test_value)
        assert " " not in h.screenshot()


def test_pymatrix_background_character_after_wake_up_typed():
    with Runner(*pymatrix_run("--test_mode", "--bg_char", ".")) as h:
        h.await_text("T")
        h.default_timeout = 10
        h.write("w")
        h.write("A")
        h.write("k")
        h.write("e")
        h.await_text("Wake up, Neo...")
        h.await_text("Knock, knock, Neo.")
        h.await_text("T")
        h.await_text(".")
        assert " " not in h.screenshot()


def test_matrix_background_character_after_wake_up_time():
    with Runner(*pymatrix_run("--test_mode",
                              "--bg_char", ".", "--wakeup")) as h:
        h.await_text("T")
        h.default_timeout = 10
        h.await_text("Wake up, Neo...")
        h.await_text("Knock, knock, Neo.")
        h.await_text("T")
        h.await_text(".")
        assert " " not in h.screenshot()


def test_matrix_background_character_reset_to_default():
    test_value = "."
    with Runner(*pymatrix_run("--bg_char", test_value,
                              "--test_mode")) as h:
        h.await_text("T")
        h.await_text(test_value)
        assert " " not in h.screenshot()
        h.press("d")
        h.await_text("T")
        h.await_text(" ")
        assert test_value not in h.screenshot()


@pytest.mark.parametrize("test_value", [
    "%r", "EE", "bg", "++", "++", "test", "3453", "__"
])
def test_pymatrix_background_character_only_one_character(test_value):
    with Runner("bash") as h:
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(". .venv/bin/activate")
        h.press("Enter")
        h.await_text("(.venv)")
        h.write(f"pymatrix-rain --test_mode --bg_char {test_value}")
        h.press("Enter")
        h.await_text(f"{test_value} is invalid. Please enter only one character")


@pytest.mark.parametrize("test_value", [
    "**", "R-", "||", "|R", "##", "test", "<<", ">>"
])
def test_pymatrix_background_character_only_one_character_quoted(test_value):
    with Runner("bash") as h:
        h.await_text("$")
        h.write("clear")
        h.press("Enter")
        h.write(". .venv/bin/activate")
        h.press("Enter")
        h.await_text("(.venv)")
        h.write(f"pymatrix-rain --test_mode --bg_char '{test_value}'")
        h.press("Enter")
        h.await_text(f"{test_value} is invalid. Please enter only one character")


def test_build_character_set_zero():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=False, katakana=False, Katakana_only=False,
        zero_one=True, test_mode=False
    )
    test_set = pymatrix.build_character_set2(args)
    assert "1" in test_set
    assert "0" in test_set
    assert "B" not in test_set
    assert "$" not in test_set
    assert "§" not in test_set
    assert "ﾎ" not in test_set


def test_build_character_set_test_char():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=False, katakana=False, Katakana_only=False,
        zero_one=False, test_mode=True
    )
    test_set = pymatrix.build_character_set2(args)
    assert test_set == ["T"]


def test_build_character_set_test_ext():
    args = pymatrix.argparse.Namespace(
        ext=True, ext_only=False, katakana=False, Katakana_only=False,
        zero_one=False, test_mode=True
    )
    test_set = pymatrix.build_character_set2(args)
    assert test_set == ["Ä", "T"]


def test_build_character_set_test_katakana_only():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=False, katakana=False, Katakana_only=True,
        zero_one=False, test_mode=True
    )
    test_set = pymatrix.build_character_set2(args)
    assert test_set == ["ﾎ", "0"]


def test_build_character_set_char_normal():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=False, katakana=False, Katakana_only=False,
        zero_one=False, test_mode=False
    )
    test_set = pymatrix.build_character_set2(args)
    assert "A" in test_set
    assert "1" in test_set
    assert "#" in test_set
    assert "a" in test_set
    assert "§" not in test_set
    assert "ﾎ" not in test_set


def test_build_character_set_ext_only():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=True, katakana=False, Katakana_only=False,
        zero_one=False, test_mode=False
    )
    test_set = pymatrix.build_character_set2(args)
    assert "§" in test_set
    assert "A" not in test_set
    assert "1" not in test_set
    assert ")" not in test_set
    assert "a" not in test_set
    assert "ﾎ" not in test_set


def test_build_character_set_test_ext_only():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=True, katakana=False, Katakana_only=False,
        zero_one=False, test_mode=True
    )
    test_set = pymatrix.build_character_set2(args)
    assert test_set == ["Ä"]


def test_build_character_set_ext_and_char():
    args = pymatrix.argparse.Namespace(
        ext=True, ext_only=False, katakana=False, Katakana_only=False,
        zero_one=False, test_mode=False
    )
    test_set = pymatrix.build_character_set2(args)
    assert "§" in test_set
    assert "A" in test_set
    assert "1" in test_set
    assert ")" in test_set
    assert "a" in test_set
    assert "ﾎ" not in test_set


def test_build_character_set_katakana_only():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=False, katakana=False, Katakana_only=True,
        zero_one=False, test_mode=False
    )
    test_set = pymatrix.build_character_set2(args)
    assert "A" not in test_set
    assert "T" not in test_set
    assert "(" not in test_set
    assert "g" not in test_set
    assert "ś" not in test_set
    assert "å" not in test_set
    assert "ｳ" in test_set
    assert "ﾒ" in test_set
    assert "1" in test_set
    assert "Z" in test_set
    assert ">" in test_set


def test_build_character_set_katakana_and_char():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=False, katakana=True, Katakana_only=False,
        zero_one=False, test_mode=False
    )
    test_set = pymatrix.build_character_set2(args)
    assert "ś" not in test_set
    assert "å" not in test_set
    assert "A" in test_set
    assert "1" in test_set
    assert "T" in test_set
    assert "(" in test_set
    assert "g" in test_set
    assert "ｳ" in test_set
    assert "ﾒ" in test_set
    assert test_set.count("0") == 1


def test_build_character_set_test_katakana_and_char():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=False, katakana=True, Katakana_only=False,
        zero_one=False, test_mode=True
    )
    test_set = pymatrix.build_character_set2(args)
    assert test_set == ["T", "ﾎ"]


def test_build_character_set_katakana_char_and_ext():
    args = pymatrix.argparse.Namespace(
        ext=True, ext_only=False, katakana=True, Katakana_only=False,
        zero_one=False, test_mode=False
    )
    test_set = pymatrix.build_character_set2(args)
    assert "A" in test_set
    assert "1" in test_set
    assert "T" in test_set
    assert "(" in test_set
    assert "g" in test_set
    assert "ś" in test_set
    assert "å" in test_set
    assert "ｳ" in test_set
    assert "ﾒ" in test_set


def test_build_character_set_test_katakana_char_and_ext():
    args = pymatrix.argparse.Namespace(
        ext=True, ext_only=False, katakana=True, Katakana_only=False,
        zero_one=False, test_mode=True
    )
    test_set = pymatrix.build_character_set2(args)
    assert test_set == ["T", "ﾎ", "Ä"]


def test_build_character_set_old_scrolling_update_list_test():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=False, katakana=False, Katakana_only=False,
        zero_one=False, test_mode=True
    )
    pymatrix.build_character_set2(args)
    assert pymatrix.OldScrollingLine.old_scroll_chr_list == ["T"]


def test_build_character_set_old_scrolling_update_list_normal():
    args = pymatrix.argparse.Namespace(
        ext=False, ext_only=False, katakana=False, Katakana_only=False,
        zero_one=False, test_mode=False
    )
    pymatrix.build_character_set2(args)
    assert "A" in pymatrix.OldScrollingLine.old_scroll_chr_list
    assert "1" in pymatrix.OldScrollingLine.old_scroll_chr_list
    assert "#" in pymatrix.OldScrollingLine.old_scroll_chr_list
    assert "a" in pymatrix.OldScrollingLine.old_scroll_chr_list
    assert "§" not in pymatrix.OldScrollingLine.old_scroll_chr_list
    assert "ﾎ" not in pymatrix.OldScrollingLine.old_scroll_chr_list


def test_list_colors_colors(capsys):
    expected = ("\033[91mred \033[92mgreen \033[94mblue \033[93myellow "
                "\033[95mmagenta \033[96mcyan \033[97mwhite \033[90mblack \033[0m\n")
    pymatrix.list_colors()
    captured_output = capsys.readouterr().out
    assert captured_output == expected


def test_pymatrix_main_list_colors(capsys):
    expected = ("\033[91mred \033[92mgreen \033[94mblue \033[93myellow "
                "\033[95mmagenta \033[96mcyan \033[97mwhite \033[90mblack \033[0m\n")
    pymatrix.main(["--list_colors"])
    captured_output = capsys.readouterr().out
    assert expected in captured_output


def test_pymatrix_main_list_commands(capsys):
    pymatrix.main(["--list_commands"])
    captured_output = capsys.readouterr().out
    assert "Commands available during run" in captured_output
    assert "Delay" in captured_output
    assert "Cycle color delay" in captured_output
